import pandas as pd
from typing import List

from mindsdb_sql_parser import ASTNode
from mindsdb.api.executor.planner.steps import FetchDataframeStepPartition
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.interfaces.query_context.context_controller import RunningQuery
from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.utilities.partitioning import get_max_thread_count, split_data_frame
from mindsdb.api.executor.sql_query.steps.fetch_dataframe import get_table_alias, get_fill_param_fnc
from mindsdb.utilities.context_executor import ContextThreadPoolExecutor


from .base import BaseStepCall


logger = log.getLogger(__name__)


class FetchDataframePartitionCall(BaseStepCall):
    """
    Alternative to FetchDataframeCall but fetch data by batches wrapping user's query to:

     select * from ({user query})
      where {track_column} > {previous value}
      order by track_column
      limit size {batch_size} `

    """

    bind = FetchDataframeStepPartition

    def call(self, step: FetchDataframeStepPartition) -> ResultSet:
        """
        Parameters:
        - batch_size - count of rows to fetch from database per iteration, optional default 1000
        - threads - run partitioning in threads, bool or int, optinal, if set:
           - int value: use this as count of threads
           - true: table threads, autodetect count of thread
           - false: disable threads even if ml task queue is enabled
        - track_column - column used for creating partitions
          - query will be sorted by this column and select will be limited by batch_size
        - error (default raise)
          - when `error='skip'`, errors in partition will be skipped and execution will be continued
        """

        self.dn = self.session.datahub.get(step.integration)
        query = step.query

        # fill params
        fill_params = get_fill_param_fnc(self.steps_data)
        query_traversal(query, fill_params)

        # get query record
        run_query = self.sql_query.run_query
        if run_query is None:
            raise RuntimeError("Error with partitioning of the query")
        run_query.set_params(step.params)

        self.table_alias = get_table_alias(step.query.from_table, self.context.get("database"))
        self.current_step_num = step.step_num
        self.substeps = step.steps

        # ml task queue enabled?
        use_threads, thread_count = False, None
        if config["ml_task_queue"]["type"] == "redis":
            use_threads = True

        # use threads?
        if "threads" in step.params:
            threads = step.params["threads"]
            if isinstance(threads, int):
                thread_count = threads
                use_threads = True
            if threads is True:
                use_threads = True
            if threads is False:
                # disable even with ml task queue
                use_threads = False

        on_error = step.params.get("error", "raise")
        if use_threads:
            return self.fetch_threads(run_query, query, thread_count=thread_count, on_error=on_error)
        else:
            return self.fetch_iterate(run_query, query, on_error=on_error)

    def fetch_iterate(self, run_query: RunningQuery, query: ASTNode, on_error: str = None) -> ResultSet:
        """
        Process batches one by one in circle
        """

        results = []

        for df in run_query.get_partitions(self.dn, self, query):
            try:
                sub_data = self.exec_sub_steps(df)
                run_query.set_progress(processed_rows=len(df))
                results.append(sub_data)
            except Exception as e:
                if on_error == "skip":
                    logger.error(e)
                else:
                    raise e

        return self.concat_results(results)

    def concat_results(self, results: List[ResultSet]) -> ResultSet:
        """
        Concatenate list of result sets to single result set
        """
        df_list = []
        for res in results:
            df, col_names = res.to_df_cols()
            if len(df) > 0:
                df_list.append(df)

        data = ResultSet()
        if len(df_list) > 0:
            data = ResultSet.from_df_cols(pd.concat(df_list), col_names)

        return data

    def exec_sub_steps(self, df: pd.DataFrame) -> ResultSet:
        """
        FetchDataframeStepPartition has substeps defined
        Every batch of data have to be used to execute these substeps
        - batch of data is put as result of FetchDataframeStepPartition
        - substep are executed using result of previos step (like it is all fetched data is available)
        - the final result is returned and used outside to concatenate with results of other's batches
        """
        input_data = ResultSet.from_df(
            df, table_name=self.table_alias[1], table_alias=self.table_alias[2], database=self.table_alias[0]
        )

        if len(self.substeps) == 0:
            return input_data

        # execute with modified previous results
        steps_data2 = self.steps_data.copy()
        steps_data2[self.current_step_num] = input_data

        sub_data = None
        for substep in self.substeps:
            sub_data = self.sql_query.execute_step(substep, steps_data=steps_data2)
            steps_data2[substep.step_num] = sub_data
        return sub_data

    def fetch_threads(
        self, run_query: RunningQuery, query: ASTNode, thread_count: int = None, on_error: str = None
    ) -> ResultSet:
        """
        Process batches in threads
        - spawn required count of threads
        - create in/out queue to communicate with threads
        - send task to threads and receive results
        """

        # create communication queues

        if thread_count is None:
            thread_count = get_max_thread_count()

        # 3 tasks per worker during 1 batch
        partition_size = int(run_query.batch_size / thread_count)
        # min partition size
        if partition_size < 10:
            partition_size = 10

        results = []

        with ContextThreadPoolExecutor(max_workers=thread_count) as executor:
            for df in run_query.get_partitions(self.dn, self, query):
                # split into chunks and send to workers
                futures = []
                for df2 in split_data_frame(df, partition_size):
                    futures.append([executor.submit(self.exec_sub_steps, df2), len(df2)])

                error = None
                for future, rows_count in futures:
                    try:
                        results.append(future.result())
                        run_query.set_progress(processed_rows=rows_count)
                    except Exception as e:
                        if on_error == "skip":
                            logger.error(e)
                        else:
                            executor.shutdown()
                            error = e

                if error:
                    raise error
                if self.sql_query.stop_event is not None and self.sql_query.stop_event.is_set():
                    executor.shutdown()
                    raise RuntimeError("Query is interrupted")

        return self.concat_results(results)



"""added data catalog tables

Revision ID: a44643042fe8
Revises: 9f150e4f9a05
Create Date: 2025-05-28 17:20:57.300313

"""

from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa
from mindsdb.interfaces.storage.db import Array


# revision identifiers, used by Alembic.
revision = "a44643042fe8"
down_revision = "9f150e4f9a05"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "meta_tables",
        sa.Column("id", sa.Integer(), primary_key=True),
        sa.Column(
            "integration_id",
            sa.Integer(),
            sa.ForeignKey("integration.id"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("schema", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("type", sa.String(), nullable=True),
        sa.Column("row_count", sa.Integer(), nullable=True),
    )

    op.create_table(
        "meta_columns",
        sa.Column("id", sa.Integer(), primary_key=True),
        sa.Column("table_id", sa.Integer(), sa.ForeignKey("meta_tables.id"), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("data_type", sa.String(), nullable=False),
        sa.Column("default_value", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("is_nullable", sa.Boolean(), nullable=True),
    )

    op.create_table(
        "meta_column_statistics",
        sa.Column(
            "column_id",
            sa.Integer(),
            sa.ForeignKey("meta_columns.id"),
            primary_key=True,
        ),
        sa.Column("most_common_values", Array(), nullable=True),
        sa.Column("most_common_frequencies", Array(), nullable=True),
        sa.Column("null_percentage", sa.Numeric(5, 2), nullable=True),
        sa.Column("distinct_values_count", sa.Integer(), nullable=True),
        sa.Column("minimum_value", sa.String(), nullable=True),
        sa.Column("maximum_value", sa.String(), nullable=True),
    )

    op.create_table(
        "meta_primary_keys",
        sa.Column("table_id", sa.Integer(), sa.ForeignKey("meta_tables.id"), primary_key=True),
        sa.Column(
            "column_id",
            sa.Integer(),
            sa.ForeignKey("meta_columns.id"),
            primary_key=True,
        ),
        sa.Column("ordinal_position", sa.Integer(), nullable=True),
        sa.Column("constraint_name", sa.String(), nullable=True),
    )

    op.create_table(
        "meta_foreign_keys",
        sa.Column(
            "parent_table_id",
            sa.Integer(),
            sa.ForeignKey("meta_tables.id"),
            primary_key=True,
        ),
        sa.Column(
            "parent_column_id",
            sa.Integer(),
            sa.ForeignKey("meta_columns.id"),
            primary_key=True,
        ),
        sa.Column(
            "child_table_id",
            sa.Integer(),
            sa.ForeignKey("meta_tables.id"),
            primary_key=True,
        ),
        sa.Column(
            "child_column_id",
            sa.Integer(),
            sa.ForeignKey("meta_columns.id"),
            primary_key=True,
        ),
        sa.Column("constraint_name", sa.String(), nullable=True),
    )


def downgrade():
    op.drop_table("meta_tables")

    op.drop_table("meta_columns")

    op.drop_table("meta_column_statistics")

    op.drop_table("meta_primary_keys")

    op.drop_table("meta_foreign_keys")



from mindsdb_sql_parser.ast import Identifier, Function, Constant, BinaryOperation


def adapt_char_fn(node: Function) -> Function | None:
    """Replace MySQL's multy-arg CHAR call to chain of DuckDB's CHR calls

    Example:
        CHAR(77, 78, 79) => CHR(77) || CHR(78) || CHR(79)

    Args:
        node (Function): Function node to adapt

    Returns:
        Function | None: Adapted function node
    """
    if len(node.args) == 1:
        node.op = "chr"
        return node

    acc = None
    for arg in node.args:
        fn = Function(op="chr", args=[arg])
        if acc is None:
            acc = fn
            continue
        acc = BinaryOperation("||", args=[acc, fn])

    acc.parentheses = True
    acc.alias = node.alias
    return acc


def adapt_locate_fn(node: Function) -> Function | None:
    """Replace MySQL's LOCATE (or INSTR) call to DuckDB's STRPOS call

    Example:
        LOCATE('bar', 'foobarbar') => STRPOS('foobarbar', 'bar')
        INSTR('foobarbar', 'bar') => STRPOS('foobarbar', 'bar')
        LOCATE('bar', 'foobarbar', 3) => ValueError (there is no analogue in DuckDB)

    Args:
        node (Function): Function node to adapt

    Returns:
        Function | None: Adapted function node

    Raises:
        ValueError: If the function has 3 arguments
    """
    if len(node.args) == 3:
        raise ValueError("MySQL LOCATE function with 3 arguments is not supported")
    if node.op == "locate":
        node.args = [node.args[1], node.args[0]]
    elif node.op == "insrt":
        node.args = [node.args[0], node.args[1]]
    node.op = "strpos"


def adapt_unhex_fn(node: Function) -> None:
    """Check MySQL's UNHEX function call arguments to ensure they are strings,
    because DuckDB's UNHEX accepts only string arguments, while MySQL's UNHEX can accept integer arguments.
    NOTE: if return dataframe from duckdb then unhex values are array - this may be an issue

    Args:
        node (Function): Function node to adapt

    Returns:
        None

    Raises:
        ValueError: If the function argument is not a string
    """
    for arg in node.args:
        if not isinstance(arg, (str, bytes)):
            raise ValueError("MySQL UNHEX function argument must be a string")


def adapt_format_fn(node: Function) -> None:
    """Adapt MySQL's FORMAT function to DuckDB's FORMAT function

    Example:
        FORMAT(1234567.89, 0) => FORMAT('{:,.0f}', 1234567.89)
        FORMAT(1234567.89, 2) => FORMAT('{:,.2f}', 1234567.89)
        FORMAT(name, 2) => FORMAT('{:,.2f}', name)
        FORMAT('{:.2f}', 1234567.89) => FORMAT('{:,.2f}', 1234567.89)  # no changes for original style

    Args:
        node (Function): Function node to adapt

    Returns:
        None

    Raises:
        ValueError: If MySQL's function has 3rd 'locale' argument, like FORMAT(12332.2, 2, 'de_DE')
    """
    match node.args[0], node.args[1]:
        case Constant(value=(int() | float())), Constant(value=int()):
            ...
        case Identifier(), Constant(value=int()):
            ...
        case _:
            return node

    if len(node.args) > 2:
        raise ValueError("'locale' argument of 'format' function is not supported")
    decimal_places = node.args[1].value

    if isinstance(node.args[0], Constant):
        node.args[1].value = node.args[0].value
        node.args[0].value = f"{{:,.{decimal_places}f}}"
    else:
        node.args[1] = node.args[0]
        node.args[0] = Constant(f"{{:,.{decimal_places}f}}")


def adapt_sha2_fn(node: Function) -> None:
    """Adapt MySQL's SHA2 function to DuckDB's SHA256 function

    Example:
        SHA2('test', 256) => SHA256('test')

    Args:
        node (Function): Function node to adapt

    Returns:
        None

    Raises:
        ValueError: If the function has more than 1 argument or the argument is not 256
    """
    if len(node.args) > 1 and node.args[1].value != 256:
        raise ValueError("Only sha256 is supported")
    node.op = "sha256"
    node.args = [node.args[0]]


def adapt_length_fn(node: Function) -> None:
    """Adapt MySQL's LENGTH function to DuckDB's STRLEN function
    NOTE: duckdb also have LENGTH, therefore it can not be used

    Example:
        LENGTH('test') => STRLEN('test')

    Args:
        node (Function): Function node to adapt

    Returns:
        None
    """
    node.op = "strlen"


def adapt_regexp_substr_fn(node: Function) -> None:
    """Adapt MySQL's REGEXP_SUBSTR function to DuckDB's REGEXP_EXTRACT function

    Example:
        REGEXP_SUBSTR('foobarbar', 'bar', 1, 1) => REGEXP_EXTRACT('foobarbar', 'bar')

    Args:
        node (Function): Function node to adapt

    Returns:
        None

    Raises:
        ValueError: If the function has more than 2 arguments or 3rd or 4th argument is not 1
    """
    if (
        len(node.args) == 3
        and node.args[2].value != 1
        or len(node.args) == 4
        and (node.args[3].value != 1 or node.args[2].value != 1)
        or len(node.args) > 4
    ):
        raise ValueError("Only 2 arguments are supported for REGEXP_SUBSTR function")
    node.args = node.args[:2]
    node.op = "regexp_extract"


def adapt_substring_index_fn(node: Function) -> BinaryOperation | Function:
    """Adapt MySQL's SUBSTRING_INDEX function to DuckDB's SPLIT_PART function

    Example:
        SUBSTRING_INDEX('a.b.c.d', '.', 1) => SPLIT_PART('a.b.c.d', '.', 1)
        SUBSTRING_INDEX('a.b.c.d', '.', 2) => CONCAT_WS('.', SPLIT_PART('a.b.c.d', '.', 1), SPLIT_PART('a.b.c.d', '.', 2))

    Args:
        node (Function): Function node to adapt

    Returns:
        BinaryOperation | Function: Binary operation node or function node

    Raises:
        ValueError: If the function has more than 3 arguments or the 3rd argument is not 1
    """
    if len(node.args[1].value) > 1:
        raise ValueError("Only one car in separator")

    if node.args[2].value == 1:
        node.op = "split_part"
        return node

    acc = [node.args[1]]
    for i in range(node.args[2].value):
        fn = Function(op="split_part", args=[node.args[0], node.args[1], Constant(i + 1)])
        acc.append(fn)

    acc = Function(op="concat_ws", args=acc)
    acc.alias = node.alias
    return acc


def adapt_curtime_fn(node: Function) -> BinaryOperation:
    """Adapt MySQL's CURTIME function to DuckDB's GET_CURRENT_TIME function.
    To get the same type as MySQL's CURTIME function, we need to cast the result to time type.

    Example:
        CURTIME() => GET_CURRENT_TIME()::time

    Args:
        node (Function): Function node to adapt

    Returns:
        BinaryOperation: Binary operation node
    """
    return BinaryOperation("::", args=[Function(op="get_current_time", args=[]), Identifier("time")], alias=node.alias)


def adapt_timestampdiff_fn(node: Function) -> None:
    """Adapt MySQL's TIMESTAMPDIFF function to DuckDB's DATE_DIFF function
    NOTE: Looks like cast string args to timestamp works in most cases, but there may be some exceptions.

    Example:
        TIMESTAMPDIFF(YEAR, '2000-02-01', '2003-05-01') => DATE_DIFF('year', timestamp '2000-02-01', timestamp '2003-05-01')

    Args:
        node (Function): Function node to adapt

    Returns:
        None
    """
    node.op = "date_diff"
    node.args[0] = Constant(node.args[0].parts[0])
    node.args[1] = BinaryOperation(" ", args=[Identifier("timestamp"), node.args[1]])
    node.args[2] = BinaryOperation(" ", args=[Identifier("timestamp"), node.args[2]])


def adapt_extract_fn(node: Function) -> None:
    """Adapt MySQL's EXTRACT function to DuckDB's EXTRACT function
    TODO: multi-part args, like YEAR_MONTH, is not supported yet
    NOTE: Looks like adding 'timestamp' works in most cases, but there may be some exceptions.

    Example:
        EXTRACT(YEAR FROM '2000-02-01') => EXTRACT('year' from timestamp '2000-02-01')

    Args:
        node (Function): Function node to adapt

    Returns:
        None
    """
    node.args[0] = Constant(node.args[0].parts[0])
    if not isinstance(node.from_arg, Identifier):
        node.from_arg = BinaryOperation(" ", args=[Identifier("timestamp"), node.from_arg])



import datetime as dt
import os
import socket
import time
from threading import Event

import sqlalchemy as sa

from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from .task_thread import TaskThread

logger = log.getLogger(__name__)


class TaskMonitor:

    MONITOR_INTERVAL_SECONDS = 2
    LOCK_EXPIRED_SECONDS = MONITOR_INTERVAL_SECONDS * 30

    def __init__(self):
        self._active_tasks = {}

    def start(self, stop_event: Event = None):
        config = Config()
        db.init()
        self.config = config

        while True:
            try:
                self.check_tasks()

                db.session.rollback()  # disable cache
                time.sleep(self.MONITOR_INTERVAL_SECONDS)

            except (SystemExit, KeyboardInterrupt):
                self.stop_all_tasks()
                return

            except Exception as e:
                logger.error(e)
                db.session.rollback()

            if stop_event is not None and stop_event.is_set():
                return

    def stop_all_tasks(self):

        active_tasks = list(self._active_tasks.keys())
        for task_id in active_tasks:
            self.stop_task(task_id)

    def check_tasks(self):
        allowed_tasks = set()

        for task in db.session.query(db.Tasks).filter(db.Tasks.active == True):  # noqa
            allowed_tasks.add(task.id)

            # start new tasks
            if task.id not in self._active_tasks:
                self.start_task(task)

        # Check active tasks
        active_tasks = list(self._active_tasks.items())
        for task_id, task in active_tasks:

            if task_id not in allowed_tasks:
                # old task
                self.stop_task(task_id)

            elif not task.is_alive():
                # dead task
                self.stop_task(task_id)

            else:
                # need to be reloaded ?
                record = db.Tasks.query.get(task_id)
                if record.reload:
                    record.reload = False
                    self.stop_task(task_id)
                else:
                    # set alive time of running tasks
                    self._set_alive(task_id)

    def _lock_task(self, task):
        run_by = f"{socket.gethostname()} {os.getpid()}"
        db_date = db.session.query(sa.func.current_timestamp()).first()[0]
        if task.run_by == run_by:
            # already locked
            task.alive_time = db_date

        elif task.alive_time is None:
            # not locked yet
            task.run_by = run_by
            task.alive_time = db_date

        elif db_date - task.alive_time > dt.timedelta(
            seconds=self.LOCK_EXPIRED_SECONDS
        ):
            # lock expired
            task.run_by = run_by
            task.alive_time = db_date

        else:
            return False

        db.session.commit()
        return True

    def _set_alive(self, task_id):
        db_date = db.session.query(sa.func.current_timestamp()).first()[0]
        task = db.Tasks.query.get(task_id)
        task.alive_time = db_date
        db.session.commit()

    def _unlock_task(self, task_id):
        task = db.Tasks.query.get(task_id)
        if task is not None:
            task.alive_time = None
            db.session.commit()

    def start_task(self, task):
        if not self._lock_task(task):
            # can't lock, skip
            return

        thread = TaskThread(task.id)

        thread.start()
        self._active_tasks[task.id] = thread

    def stop_task(self, task_id: int):
        thread = self._active_tasks[task_id]
        thread.stop()
        thread.join(1)

        if thread.is_alive():
            # don't delete task, wait next circle
            return

        del self._active_tasks[task_id]
        self._unlock_task(task_id)


def start(verbose=False):

    monitor = TaskMonitor()
    monitor.start()


if __name__ == "__main__":
    start()



from http import HTTPStatus

from flask import request
from flask_restx import Resource
from langchain_text_splitters import MarkdownHeaderTextSplitter

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.exceptions import ExecutorException
from mindsdb.api.http.utils import http_error

from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.integrations.utilities.rag.splitters.file_splitter import FileSplitter, FileSplitterConfig
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.knowledge_base.preprocessing.constants import (
    DEFAULT_CONTEXT_DOCUMENT_LIMIT,
    DEFAULT_CRAWL_DEPTH,
    DEFAULT_WEB_FILTERS,
    DEFAULT_MARKDOWN_HEADERS,
    DEFAULT_WEB_CRAWL_LIMIT,
)
from mindsdb.interfaces.knowledge_base.preprocessing.document_loader import DocumentLoader
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseTable
from mindsdb.utilities import log
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.integrations.utilities.rag.settings import DEFAULT_LLM_MODEL, DEFAULT_RAG_PROMPT_TEMPLATE


from mindsdb_sql_parser.ast import Identifier

logger = log.getLogger(__name__)


@ns_conf.route("/<project_name>/knowledge_bases")
class KnowledgeBasesResource(Resource):
    @ns_conf.doc("list_knowledge_bases")
    @api_endpoint_metrics("GET", "/knowledge_bases")
    def get(self, project_name):
        """List all knowledge bases"""
        session = SessionController()
        project_controller = ProjectController()
        try:
            _ = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        # KB Controller already returns dict.
        return session.kb_controller.list(project_name)

    @ns_conf.doc("create_knowledge_base")
    @api_endpoint_metrics("POST", "/knowledge_bases")
    def post(self, project_name):
        """Create a knowledge base"""

        # Check for required parameters.
        if "knowledge_base" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "knowledge_base" parameter in POST body'
            )

        knowledge_base = request.json["knowledge_base"]
        # Explicitly require embedding model & vector database.
        required_fields = ["name"]
        for field in required_fields:
            if field not in knowledge_base:
                return http_error(
                    HTTPStatus.BAD_REQUEST, "Missing parameter", f'Must provide "{field}" field in "knowledge_base"'
                )
        if "storage" in knowledge_base:
            if "database" not in knowledge_base["storage"] or "table" not in knowledge_base["storage"]:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    "Missing parameter",
                    'Must provide "database" and "table" field in "storage" param',
                )

        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        kb_name = knowledge_base.get("name")
        existing_kb = session.kb_controller.get(kb_name, project.id)
        if existing_kb is not None:
            # Knowledge Base must not exist.
            return http_error(
                HTTPStatus.CONFLICT,
                "Knowledge Base already exists",
                f"Knowledge Base with name {kb_name} already exists",
            )

        # Legacy: Support for embedding model identifier.
        # embedding_model_identifier = None
        # if knowledge_base.get('model'):
        #     embedding_model_identifier = Identifier(parts=[knowledge_base['model']])

        storage = knowledge_base.get("storage")
        embedding_table_identifier = None
        if storage is not None:
            embedding_table_identifier = Identifier(parts=[storage["database"], storage["table"]])

        params = knowledge_base.get("params", {})

        optional_parameter_fields = [
            "embedding_model",
            "reranking_model",
            "content_columns",
            "metadata_columns",
            "id_column",
        ]

        for field in optional_parameter_fields:
            if field in knowledge_base:
                params[field] = knowledge_base[field]

        try:
            new_kb = session.kb_controller.add(
                kb_name,
                project.name,
                embedding_table_identifier,
                params=params,
                preprocessing_config=knowledge_base.get("preprocessing"),
            )
        except ValueError as e:
            return http_error(HTTPStatus.BAD_REQUEST, "Invalid preprocessing configuration", str(e))

        return new_kb.as_dict(session.show_secrets), HTTPStatus.CREATED


@ns_conf.route("/<project_name>/knowledge_bases/<knowledge_base_name>")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("knowledge_base_name", "Name of the knowledge_base")
class KnowledgeBaseResource(Resource):
    @ns_conf.doc("get_knowledge_base")
    @api_endpoint_metrics("GET", "/knowledge_bases/knowledge_base")
    def get(self, project_name, knowledge_base_name):
        """Gets a knowledge base by name"""
        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        existing_kb = session.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                "Knowledge Base not found",
                f"Knowledge Base with name {knowledge_base_name} does not exist",
            )
        return existing_kb.as_dict(session.show_secrets), HTTPStatus.OK

    @ns_conf.doc("update_knowledge_base")
    @api_endpoint_metrics("PUT", "/knowledge_bases/knowledge_base")
    def put(self, project_name: str, knowledge_base_name: str):
        """Updates a knowledge base with optional preprocessing."""

        # Check for required parameters
        if "knowledge_base" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "knowledge_base" parameter in PUT body'
            )

        session = SessionController()
        project_controller = ProjectController()

        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        try:
            kb_data = request.json["knowledge_base"]

            # Retrieve the knowledge base table for updates
            table = session.kb_controller.get_table(knowledge_base_name, project.id, params=kb_data.get("params"))
            if table is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    "Knowledge Base not found",
                    f"Knowledge Base with name {knowledge_base_name} does not exist",
                )

            # Set up dependencies for DocumentLoader
            file_controller = FileController()
            file_splitter_config = FileSplitterConfig()
            file_splitter = FileSplitter(file_splitter_config)
            markdown_splitter = MarkdownHeaderTextSplitter(headers_to_split_on=DEFAULT_MARKDOWN_HEADERS)
            mysql_proxy = FakeMysqlProxy()

            # Initialize DocumentLoader with required components
            document_loader = DocumentLoader(
                file_controller=file_controller,
                file_splitter=file_splitter,
                markdown_splitter=markdown_splitter,
                mysql_proxy=mysql_proxy,
            )

            # Configure table with dependencies
            table.document_loader = document_loader

            # Update preprocessing configuration if provided
            if "preprocessing" in kb_data:
                table.configure_preprocessing(kb_data["preprocessing"])

            # Process raw data rows if provided
            if kb_data.get("rows"):
                table.insert_rows(kb_data["rows"])

            # Process files if specified
            if kb_data.get("files"):
                table.insert_files(kb_data["files"])

            # Process web pages if URLs provided
            if kb_data.get("urls"):
                table.insert_web_pages(
                    urls=kb_data["urls"],
                    limit=kb_data.get("limit") or DEFAULT_WEB_CRAWL_LIMIT,
                    crawl_depth=kb_data.get("crawl_depth", DEFAULT_CRAWL_DEPTH),
                    filters=kb_data.get("filters", DEFAULT_WEB_FILTERS),
                )

            # Process query if provided
            if kb_data.get("query"):
                table.insert_query_result(kb_data["query"], project_name)

        except ExecutorException as e:
            logger.error(f"Error during preprocessing and insertion: {str(e)}")
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Invalid SELECT query",
                f'Executing "query" failed. Needs to be a valid SELECT statement that returns data: {str(e)}',
            )

        except Exception as e:
            logger.error(f"Error during preprocessing and insertion: {str(e)}")
            return http_error(
                HTTPStatus.BAD_REQUEST, "Preprocessing Error", f"Error during preprocessing and insertion: {str(e)}"
            )

        return "", HTTPStatus.OK

    @ns_conf.doc("delete_knowledge_base")
    @api_endpoint_metrics("DELETE", "/knowledge_bases/knowledge_base")
    def delete(self, project_name: str, knowledge_base_name: str):
        """Deletes a knowledge base."""
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        session_controller = SessionController()
        existing_kb = session_controller.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            # Knowledge Base must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                "Knowledge Base not found",
                f"Knowledge Base with name {knowledge_base_name} does not exist",
            )

        session_controller.kb_controller.delete(knowledge_base_name, project_name)
        return "", HTTPStatus.NO_CONTENT


def _handle_chat_completion(knowledge_base_table: KnowledgeBaseTable, request):
    # Check for required parameters
    query = request.json.get("query")

    llm_model = request.json.get("llm_model")
    if llm_model is None:
        logger.warning(f'Missing parameter "llm_model" in POST body, using default llm_model {DEFAULT_LLM_MODEL}')

    prompt_template = request.json.get("prompt_template")
    if prompt_template is None:
        logger.warning(
            f'Missing parameter "prompt_template" in POST body, using default prompt template {DEFAULT_RAG_PROMPT_TEMPLATE}'
        )

    # Get retrieval config, if set
    retrieval_config = request.json.get("retrieval_config", {})
    if not retrieval_config:
        logger.warning("No retrieval config provided, using default retrieval config")

    # add llm model to retrieval config
    if llm_model is not None:
        retrieval_config["llm_model_name"] = llm_model

    # add prompt template to retrieval config
    if prompt_template is not None:
        retrieval_config["rag_prompt_template"] = prompt_template

    # add llm provider to retrieval config if set
    llm_provider = request.json.get("model_provider")
    if llm_provider is not None:
        retrieval_config["llm_provider"] = llm_provider

    # build rag pipeline
    rag_pipeline = knowledge_base_table.build_rag_pipeline(retrieval_config)

    # get response from rag pipeline
    rag_response = rag_pipeline(query)
    response = {
        "message": {"content": rag_response.get("answer"), "context": rag_response.get("context"), "role": "assistant"}
    }

    return response


def _handle_context_completion(knowledge_base_table: KnowledgeBaseTable, request):
    # Used for semantic search.
    query = request.json.get("query")
    # Keyword search.
    keywords = request.json.get("keywords")
    # Metadata search.
    metadata = request.json.get("metadata")
    # Maximum amount of documents to return as context.
    limit = request.json.get("limit", DEFAULT_CONTEXT_DOCUMENT_LIMIT)

    # Use default distance function & column names for ID, content, & metadata, to keep things simple.
    hybrid_search_df = knowledge_base_table.hybrid_search(query, keywords=keywords, metadata=metadata)

    num_documents = len(hybrid_search_df.index)
    context_documents = []
    for i in range(limit):
        if i >= num_documents:
            break
        row = hybrid_search_df.iloc[i]
        context_documents.append({"id": row["id"], "content": row["content"], "rank": row["rank"]})

    return {"documents": context_documents}


@ns_conf.route("/<project_name>/knowledge_bases/<knowledge_base_name>/completions")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("knowledge_base_name", "Name of the knowledge_base")
class KnowledgeBaseCompletions(Resource):
    @ns_conf.doc("knowledge_base_completions")
    @api_endpoint_metrics("POST", "/knowledge_bases/knowledge_base/completions")
    def post(self, project_name, knowledge_base_name):
        """
        Add support for LLM generation on the response from knowledge base. Default completion type is 'chat' unless specified.
        """
        if request.json.get("query") is None:
            # "query" is used for semantic search for both completion types.
            logger.error('Missing parameter "query" in POST body')
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "query" parameter in POST body'
            )

        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            logger.error("Project not found, please check the project name exists")
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        session = SessionController()
        # Check if knowledge base exists
        table = session.kb_controller.get_table(knowledge_base_name, project.id)
        if table is None:
            logger.error("Knowledge Base not found, please check the knowledge base name exists")
            return http_error(
                HTTPStatus.NOT_FOUND,
                "Knowledge Base not found",
                f"Knowledge Base with name {knowledge_base_name} does not exist",
            )

        completion_type = request.json.get("type", "chat")
        if completion_type == "context":
            return _handle_context_completion(table, request)
        if completion_type == "chat":
            return _handle_chat_completion(table, request)
        return http_error(
            HTTPStatus.BAD_REQUEST,
            "Invalid parameter",
            f'Completion type must be one of: "context", "chat". Received {completion_type}',
        )



import os
import textwrap
from typing import Optional, Dict, Any
import requests
import pandas as pd
from openai import OpenAI, AuthenticationError
from mindsdb.integrations.handlers.openai_handler import Handler as OpenAIHandler
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.handlers.togetherai_handler.settings import (
    togetherai_handler_config,
)

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class TogetherAIHandler(OpenAIHandler):
    """
    This handler handles connection to the TogetherAI.
    """

    name = "togetherai"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.api_base = togetherai_handler_config.BASE_URL
        self.default_model = togetherai_handler_config.DEFAULT_MODEL
        self.default_embedding_model = togetherai_handler_config.DEFAULT_EMBEDDING_MODEL
        self.default_mode = togetherai_handler_config.DEFAULT_MODE
        self.supported_modes = togetherai_handler_config.SUPPORTED_MODES

    @staticmethod
    def _check_client_connection(client: OpenAI):
        """
        Check the TogetherAI engine client connection by listing models.

        Args:
            client (OpenAI): OpenAI client configured with the TogetherAI API credentials.

        Raises:
            Exception: If the client connection (API key) is invalid.

        Returns:
            None
        """

        try:
            TogetherAIHandler._get_supported_models(client.api_key, client.base_url)

        except Exception as e:
            raise Exception(f"Something went wrong: {e}")

    def create_engine(self, connection_args):
        """
        Validate the TogetherAI API credentials on engine creation.

        Args:
            connection_args (dict): Connection arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """

        connection_args = {k.lower(): v for k, v in connection_args.items()}
        api_key = connection_args.get("togetherai_api_key")
        if api_key is not None:
            api_base = connection_args.get("api_base") or os.environ.get(
                "TOGETHERAI_API_BASE", togetherai_handler_config.BASE_URL
            )
            client = self._get_client(api_key=api_key, base_url=api_base)
            TogetherAIHandler._check_client_connection(client)

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        """
        Validate the TogetherAI API credentials on model creation.

        Args:
            target (str): Target column, not required for LLMs.
            args (dict): Handler arguments.
            kwargs (dict): Handler keyword arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """
        if "using" not in args:
            raise Exception(
                "TogetherAI engine require a USING clause! Refer to its documentation for more details"
            )
        else:
            args = args["using"]

        if (
            len(set(args.keys()) & {"question_column", "prompt_template", "prompt"})
            == 0
        ):
            raise Exception(
                "One of `question_column`, `prompt_template` or `prompt` is required for this engine."
            )

        keys_collection = [
            ["prompt_template"],
            ["question_column", "context_column"],
            ["prompt", "user_column", "assistant_column"],
        ]
        for keys in keys_collection:
            if keys[0] in args and any(
                x[0] in args for x in keys_collection if x != keys
            ):
                raise Exception(
                    textwrap.dedent(
                        """\
                    Please provide one of
                        1) a `prompt_template`
                        2) a `question_column` and an optional `context_column`
                        3) a `prompt`, `user_column` and `assistant_column`
                """
                    )
                )

        engine_storage = kwargs["handler_storage"]
        connection_args = engine_storage.get_connection_args()
        api_key = get_api_key("togetherai", args, engine_storage=engine_storage)
        api_base = connection_args.get("api_base") or os.environ.get(
            "TOGETHERAI_API_BASE", togetherai_handler_config.BASE_URL
        )
        client = TogetherAIHandler._get_client(api_key=api_key, base_url=api_base)
        TogetherAIHandler._check_client_connection(client)

    def create(self, target, args: Dict = None, **kwargs: Any) -> None:
        """
        Create a model for TogetherAI engine.

        Args:
            target (str): Target column, not required for LLMs.
            args (dict): Handler arguments.
            kwargs (dict): Handler keyword arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """
        args = args["using"]
        args["target"] = target
        try:
            api_key = get_api_key(self.api_key_name, args, self.engine_storage)
            connection_args = self.engine_storage.get_connection_args()
            api_base = (
                args.get("api_base")
                or connection_args.get("api_base")
                or os.environ.get("TOGETHERAI_API_BASE")
                or self.api_base
            )
            available_models = self._get_supported_models(api_key, api_base)

            if args.get("mode") is None:
                args["mode"] = self.default_mode
            elif args["mode"] not in self.supported_modes:
                raise Exception(
                    f"Invalid operation mode. Please use one of {self.supported_modes}"
                )

            if not args.get("model_name"):
                if args["mode"] == "embedding":
                    args["model_name"] = self.default_embedding_model
                else:
                    args["model_name"] = self.default_model
            elif args["model_name"] not in available_models:
                raise Exception(
                    f"Invalid model name. Please use one of {available_models}"
                )
        finally:
            self.model_storage.json_set("args", args)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Call the TogetherAI engine to predict the next token.

        Args:
            df (pd.DataFrame): Input data.
            args (dict): Handler arguments.

        Returns:
            pd.DataFrame: Predicted data.
        """

        api_key = get_api_key("togetherai", args, engine_storage=self.engine_storage)
        supported_models = self._get_supported_models(api_key, self.api_base)
        self.chat_completion_models = supported_models
        return super().predict(df, args)

    @staticmethod
    def _get_supported_models(api_key, base_url):
        """
        Get the list of supported models from the TogetherAI engine.

        Args:
            api_key (str): TogetherAI API key.
            base_url (str): TogetherAI API base URL.

        Returns:
            list: List of supported models.
        """

        list_model_endpoint = f"{base_url}/models"
        headers = {
            "accept": "application/json",
            "authorization": f"Bearer {api_key}",
        }
        response = requests.get(url=list_model_endpoint, headers=headers)

        if response.status_code == 200:
            model_list = response.json()
            chat_completion_models = list(map(lambda model: model["id"], model_list))
            return chat_completion_models
        elif response.status_code == 401:
            raise AuthenticationError(message="Invalid API key")
        else:
            raise Exception(f"Failed to get supported models: {response.text}")

    def finetune(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> None:
        raise NotImplementedError("Fine-tuning is not supported for TogetherAI engine")



from mindsdb.interfaces.data_catalog.base_data_catalog import BaseDataCatalog
from mindsdb.interfaces.storage import db


class DataCatalogReader(BaseDataCatalog):
    """
    This class is responsible for reading the metadata from the data catalog and providing it in a structured format.
    """

    def read_metadata_as_string(self) -> str:
        """
        Read the metadata from the data catalog and return it as a string.
        """
        tables = self._read_metadata()
        if not tables:
            self.logger.warning(f"No metadata found for database '{self.database_name}'")
            return f"No metadata found for database '{self.database_name}'"

        metadata_str = "Data Catalog: \n"
        if hasattr(self.data_handler, "meta_get_handler_info"):
            info = self.data_handler.meta_get_handler_info()
            if info:
                metadata_str += info + "\n\n"

        for table in tables:
            metadata_str += table.as_string() + "\n\n"
        return metadata_str

    def read_metadata_as_records(self) -> list:
        """
        Read the metadata from the data catalog and return it as a list of database records.
        """
        tables = self._read_metadata()
        if not tables:
            self.logger.warning(f"No metadata found for database '{self.database_name}'")
            return []
        return tables

    def get_handler_info(self) -> str:
        """
        Get the handler info for the database.
        """
        return self.data_handler.meta_get_handler_info()

    def _read_metadata(self) -> list:
        """
        Read the metadata from the data catalog and return it in a structured format.
        """
        if not self.is_data_catalog_supported():
            return f"Data catalog is not supported for database '{self.database_name}'."

        query = db.session.query(db.MetaTables).filter_by(integration_id=self.integration_id)
        if self.table_names:
            cleaned_table_names = [name.strip("`").split(".")[-1] for name in self.table_names]
            query = query.filter(db.MetaTables.name.in_(cleaned_table_names))
        tables = query.all()
        return tables



"""queries

Revision ID: fda503400e43
Revises: 11347c213b36
Create Date: 2025-03-21 18:50:20.795930

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = 'fda503400e43'
down_revision = '11347c213b36'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'queries',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('sql', sa.String(), nullable=False),
        sa.Column('started_at', sa.DateTime(), nullable=True),
        sa.Column('finished_at', sa.DateTime(), nullable=True),
        sa.Column('parameters', sa.JSON(), nullable=True),
        sa.Column('context', sa.JSON(), nullable=True),
        sa.Column('processed_rows', sa.Integer(), nullable=True),
        sa.Column('error', sa.String(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    with op.batch_alter_table('knowledge_base', schema=None) as batch_op:
        batch_op.add_column(sa.Column('query_id', sa.INTEGER(), nullable=True))


def downgrade():
    with op.batch_alter_table('knowledge_base', schema=None) as batch_op:
        batch_op.drop_column('query_id')

    op.drop_table('queries')



import copy

from mindsdb_sql_parser.ast import (
    Identifier,
    Constant,
    Delete,
    Parameter,
    Tuple,
)
from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.api.executor.planner.steps import DeleteStep

from mindsdb.api.executor.sql_query.result_set import ResultSet

from .base import BaseStepCall


class DeleteStepCall(BaseStepCall):

    bind = DeleteStep

    def call(self, step):
        if len(step.table.parts) > 1:
            integration_name = step.table.parts[0]
            table_name_parts = step.table.parts[1:]
        else:
            integration_name = self.context['database']
            table_name_parts = step.table.parts

        dn = self.session.datahub.get(integration_name)

        # make command
        query = Delete(
            table=Identifier(parts=table_name_parts),
            where=copy.deepcopy(step.where),
        )

        # fill params
        def fill_params(node, **kwargs):
            if isinstance(node, Parameter):
                rs = self.steps_data[node.value.step_num]
                items = [Constant(i) for i in rs.get_column_values(col_idx=0)]
                return Tuple(items)

        query_traversal(query.where, fill_params)

        response = dn.query(query=query, session=self.session)
        return ResultSet(affected_rows=response.affected_rows)



__title__ = "MindsDB"
__package_name__ = "mindsdb"
__version__ = "25.9.1.0"
__description__ = "MindsDB's AI SQL Server enables developers to build AI tools that need access to real-time data to perform their tasks"
__email__ = "jorge@mindsdb.com"
__author__ = "MindsDB Inc"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "Elastic License 2.0"
__copyright__ = "Copyright(c) 2018 MindsDB, Inc"



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.STR,
        "description": "Your PirateWeather API key.",
        "secret": True
    }
)

connection_args_example = OrderedDict(api_key="knlsndlknslk")



import praw
import os
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log

from .reddit_tables import CommentTable, SubmissionTable

logger = log.getLogger(__name__)


class RedditHandler(APIHandler):

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get('connection_data', {})

        self.connection_args = {}
        handler_config = Config().get('reddit_handler', {})
        for k in ['client_id', 'client_secret', 'user_agent']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'REDDIT_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'REDDIT_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.reddit = None
        self.is_connected = False

        comment = CommentTable(self)
        self._register_table('comment', comment)

        submission = SubmissionTable(self)
        self._register_table('submission', submission)

    def connect(self):
        """Authenticate with the Reddit API using the client ID, client secret
        and user agent provided in the constructor.
        """
        if self.is_connected is True:
            return self.reddit

        self.reddit = praw.Reddit(
            client_id=self.connection_args['client_id'],
            client_secret=self.connection_args['client_secret'],
            user_agent=self.connection_args['user_agent'],
        )

        self.is_connected = True
        return self.reddit

    def check_connection(self) -> StatusResponse:
        '''It evaluates if the connection with Reddit API is alive and healthy.
        Returns:
            HandlerStatusResponse
        '''

        response = StatusResponse(False)

        try:
            reddit = self.connect()
            reddit.user.me()
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Reddit api: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None):
        '''It parses any native statement string and acts upon it (for example, raw syntax commands).
        Args:
        query (Any): query in native format (str for sql databases,
            dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        '''

        method_name, params = self.parse_native_query(query_string)
        if method_name == 'get_submission':
            df = self.get_submission(params)
        elif method_name == 'get_subreddit':
            df = self.get_subreddit(params)
        else:
            raise ValueError(f"Method '{method_name}' not supported by RedditHandler")

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'User name',
        'required': True,
        'label': 'User',
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Password',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    protocol_version={
        'type': ARG_TYPE.INT,
        'description': 'is not required and defaults to 4.',
        'required': False,
        'label': 'Protocol version',
    },
    host={
        'type': ARG_TYPE.STR,
        'description': ' is the host name or IP address of the Cassandra database.',
        'required': True,
        'label': 'Host',
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Server port',
        'required': True,
        'label': 'Port',
    },
    keyspace={
        'type': ARG_TYPE.STR,
        'description': ' is the keyspace to connect, the top level container for tables.',
        'required': True,
        'label': 'Keyspace',
    },
    secure_connect_bundle={
        'type': ARG_TYPE.STR,
        'description': 'Path or URL to the secure connect bundle',
        'required': False,
        'label': 'Host',
    },
)



from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter
from opentelemetry.sdk.trace.sampling import Sampler


def setup_tracer(resource: Resource, sampler: Sampler, exporter: SpanExporter) -> None:
    """
    Setup OpenTelemetry tracing
    """
    # Set the tracer provider with the custom resource
    trace.set_tracer_provider(TracerProvider(resource=resource, sampler=sampler))

    # Replace the default span processor with the custom one
    trace.get_tracer_provider().add_span_processor(BatchSpanProcessor(exporter))



import os
import copy
from typing import Dict, List, Optional, Any, Text
import json
import decimal

import pandas as pd
import numpy as np
from pydantic import BaseModel, ValidationError
from sqlalchemy.orm.attributes import flag_modified

from mindsdb_sql_parser.ast import BinaryOperation, Constant, Identifier, Select, Update, Delete, Star
from mindsdb_sql_parser.ast.mindsdb import CreatePredictor
from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.libs.keyword_search_base import KeywordSearchBase
from mindsdb.integrations.utilities.query_traversal import query_traversal

import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.vectordatabase_handler import (
    DistanceFunction,
    TableField,
    VectorStoreHandler,
)
from mindsdb.integrations.utilities.rag.rag_pipeline_builder import RAG
from mindsdb.integrations.utilities.rag.config_loader import load_rag_config
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import (
    construct_model_from_args,
)

from mindsdb.interfaces.agents.constants import DEFAULT_EMBEDDINGS_MODEL_CLASS, MAX_INSERT_BATCH_SIZE
from mindsdb.interfaces.agents.langchain_agent import create_chat_model, get_llm_provider
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.variables.variables_controller import variables_controller
from mindsdb.interfaces.knowledge_base.preprocessing.models import PreprocessingConfig, Document
from mindsdb.interfaces.knowledge_base.preprocessing.document_preprocessor import PreprocessorFactory
from mindsdb.interfaces.knowledge_base.evaluate import EvaluateBase
from mindsdb.interfaces.knowledge_base.executor import KnowledgeBaseQueryExecutor
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator, KeywordSearchArgs
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx

from mindsdb.api.executor.command_executor import ExecuteCommands
from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.utilities import log
from mindsdb.integrations.utilities.rag.rerankers.base_reranker import BaseLLMReranker

logger = log.getLogger(__name__)


class KnowledgeBaseInputParams(BaseModel):
    metadata_columns: List[str] | None = None
    content_columns: List[str] | None = None
    id_column: str | None = None
    kb_no_upsert: bool = False
    embedding_model: Dict[Text, Any] | None = None
    is_sparse: bool = False
    vector_size: int | None = None
    reranking_model: Dict[Text, Any] | None = None
    preprocessing: Dict[Text, Any] | None = None

    class Config:
        extra = "forbid"


def get_model_params(model_params: dict, default_config_key: str):
    """
    Get model parameters by combining default config with user provided parameters.
    """
    combined_model_params = copy.deepcopy(config.get(default_config_key, {}))

    if model_params:
        if not isinstance(model_params, dict):
            raise ValueError("Model parameters must be passed as a JSON object")

        combined_model_params.update(model_params)

    combined_model_params.pop("use_default_llm", None)

    return combined_model_params


def get_embedding_model_from_params(embedding_model_params: dict):
    """
    Create embedding model from parameters.
    """
    params_copy = copy.deepcopy(embedding_model_params)
    provider = params_copy.pop("provider", None).lower()
    api_key = get_api_key(provider, params_copy, strict=False) or params_copy.get("api_key")
    # Underscores are replaced because the provider name ultimately gets mapped to a class name.
    # This is mostly to support Azure OpenAI (azure_openai); the mapped class name is 'AzureOpenAIEmbeddings'.
    params_copy["class"] = provider.replace("_", "")
    if provider == "azure_openai":
        # Azure OpenAI expects the api_key to be passed as 'openai_api_key'.
        params_copy["openai_api_key"] = api_key
        params_copy["azure_endpoint"] = params_copy.pop("base_url")
        if "chunk_size" not in params_copy:
            params_copy["chunk_size"] = 2048
        if "api_version" in params_copy:
            params_copy["openai_api_version"] = params_copy["api_version"]
    else:
        params_copy[f"{provider}_api_key"] = api_key
    params_copy.pop("api_key", None)
    params_copy["model"] = params_copy.pop("model_name", None)

    return construct_model_from_args(params_copy)


def get_reranking_model_from_params(reranking_model_params: dict):
    """
    Create reranking model from parameters.
    """
    params_copy = copy.deepcopy(reranking_model_params)
    provider = params_copy.get("provider", "openai").lower()

    if "api_key" not in params_copy:
        params_copy["api_key"] = get_api_key(provider, params_copy, strict=False)

    if "model_name" not in params_copy:
        raise ValueError("'model_name' must be provided for reranking model")
    params_copy["model"] = params_copy.pop("model_name")

    return BaseLLMReranker(**params_copy)


def safe_pandas_is_datetime(value: str) -> bool:
    """
    Check if the value can be parsed as a datetime.
    """
    try:
        result = pd.api.types.is_datetime64_any_dtype(value)
        return result
    except ValueError:
        return False


def to_json(obj):
    if obj is None:
        return None
    try:
        return json.dumps(obj)
    except TypeError:
        return obj


class KnowledgeBaseTable:
    """
    Knowledge base table interface
    Handlers requests to KB table and modifies data in linked vector db table
    """

    def __init__(self, kb: db.KnowledgeBase, session):
        self._kb = kb
        self._vector_db = None
        self.session = session
        self.document_preprocessor = None
        self.document_loader = None
        self.model_params = None

        self.kb_to_vector_columns = {"id": "_original_doc_id", "chunk_id": "id", "chunk_content": "content"}
        if self._kb.params.get("version", 0) < 2:
            self.kb_to_vector_columns["id"] = "original_doc_id"

    def configure_preprocessing(self, config: Optional[dict] = None):
        """Configure preprocessing for the knowledge base table"""
        logger.debug(f"Configuring preprocessing with config: {config}")
        self.document_preprocessor = None  # Reset existing preprocessor
        if config is None:
            config = {}

        # Ensure content_column is set for JSON chunking if not already specified
        if config.get("type") == "json_chunking" and config.get("json_chunking_config"):
            if "content_column" not in config["json_chunking_config"]:
                config["json_chunking_config"]["content_column"] = "content"

        preprocessing_config = PreprocessingConfig(**config)
        self.document_preprocessor = PreprocessorFactory.create_preprocessor(preprocessing_config)

        # set doc_id column name
        self.document_preprocessor.config.doc_id_column_name = self.kb_to_vector_columns["id"]

        logger.debug(f"Created preprocessor of type: {type(self.document_preprocessor)}")

    def select_query(self, query: Select) -> pd.DataFrame:
        """
        Handles select from KB table.
        Replaces content values with embeddings in where clause. Sends query to vector db
        :param query: query to KB table
        :return: dataframe with the result table
        """

        # Copy query for complex execution via DuckDB: DISTINCT, GROUP BY etc.
        query_copy = copy.deepcopy(query)

        executor = KnowledgeBaseQueryExecutor(self)
        df = executor.run(query)

        if (
            query_copy.group_by is not None
            or query_copy.order_by is not None
            or query_copy.having is not None
            or query_copy.distinct is True
            or len(query_copy.targets) != 1
            or not isinstance(query_copy.targets[0], Star)
        ):
            query_copy.where = None
            if "metadata" in df.columns:
                df["metadata"] = df["metadata"].apply(to_json)

            if query_copy.from_table is None:
                query_copy.from_table = Identifier(parts=[self._kb.name])

            df = query_df(df, query_copy, session=self.session)

        return df

    def select(self, query, disable_reranking=False):
        logger.debug(f"Processing select query: {query}")

        # Extract the content query text for potential reranking

        db_handler = self.get_vector_db()

        logger.debug("Replaced content with embeddings in where clause")
        # set table name
        query.from_table = Identifier(parts=[self._kb.vector_database_table])
        logger.debug(f"Set table name to: {self._kb.vector_database_table}")

        query.targets = [
            Identifier(TableField.ID.value),
            Identifier(TableField.CONTENT.value),
            Identifier(TableField.METADATA.value),
            Identifier(TableField.DISTANCE.value),
        ]

        # Get response from vector db
        logger.debug(f"Using vector db handler: {type(db_handler)}")

        # extract values from conditions and prepare for vectordb
        conditions = []
        keyword_search_conditions = []
        keyword_search_cols_and_values = []
        query_text = None
        relevance_threshold = None
        hybrid_search_enabled_flag = False
        query_conditions = db_handler.extract_conditions(query.where)
        hybrid_search_alpha = None  # Default to None, meaning no alpha weighted blending
        if query_conditions is not None:
            for item in query_conditions:
                if item.column == "relevance" and item.op.value == FilterOperator.GREATER_THAN_OR_EQUAL.value:
                    try:
                        relevance_threshold = float(item.value)
                        # Validate range: must be between 0 and 1
                        if not (0 <= relevance_threshold <= 1):
                            raise ValueError(f"relevance_threshold must be between 0 and 1, got: {relevance_threshold}")
                        logger.debug(f"Found relevance_threshold in query: {relevance_threshold}")
                    except (ValueError, TypeError) as e:
                        error_msg = f"Invalid relevance_threshold value: {item.value}. {str(e)}"
                        logger.error(error_msg)
                        raise ValueError(error_msg)
                elif item.column == "reranking":
                    if item.value is False or (isinstance(item.value, str) and item.value.lower() == "false"):
                        disable_reranking = True
                elif item.column == "hybrid_search":
                    hybrid_search_enabled_flag = item.value
                    # cast to boolean
                    if isinstance(hybrid_search_enabled_flag, str):
                        hybrid_search_enabled_flag = hybrid_search_enabled_flag.lower() not in ("false")
                    if item.value is False or (isinstance(item.value, str) and item.value.lower() == "false"):
                        disable_reranking = True
                elif item.column == "hybrid_search_alpha":
                    # validate item.value is a float
                    if not isinstance(item.value, (float, int)):
                        raise ValueError(f"Invalid hybrid_search_alpha value: {item.value}. Must be a float or int.")
                    # validate hybrid search alpha is between 0 and 1
                    if not (0 <= item.value <= 1):
                        raise ValueError(f"Invalid hybrid_search_alpha value: {item.value}. Must be between 0 and 1.")
                    hybrid_search_alpha = item.value
                elif item.column == "relevance" and item.op.value != FilterOperator.GREATER_THAN_OR_EQUAL.value:
                    raise ValueError(
                        f"Invalid operator for relevance: {item.op.value}. Only GREATER_THAN_OR_EQUAL is allowed."
                    )
                elif item.column == TableField.CONTENT.value:
                    query_text = item.value

                    # replace content with embeddings
                    conditions.append(
                        FilterCondition(
                            column=TableField.EMBEDDINGS.value,
                            value=self._content_to_embeddings(item.value),
                            op=FilterOperator.EQUAL,
                        )
                    )
                    keyword_search_cols_and_values.append((TableField.CONTENT.value, item.value))
                else:
                    conditions.append(item)
                    keyword_search_conditions.append(item)  # keyword search conditions do not use embeddings

        if len(keyword_search_cols_and_values) > 1:
            raise ValueError(
                "Multiple content columns found in query conditions. "
                "Only one content column is allowed for keyword search."
            )

        logger.debug(f"Extracted query text: {query_text}")

        self.addapt_conditions_columns(conditions)

        # Set default limit if query is present
        if query_text is not None:
            limit = query.limit.value if query.limit is not None else None
            if limit is None:
                limit = 10
            elif limit > 100:
                limit = 100
            query.limit = Constant(limit)

        allowed_metadata_columns = self._get_allowed_metadata_columns()
        df = db_handler.dispatch_select(query, conditions, allowed_metadata_columns=allowed_metadata_columns)
        df = self.addapt_result_columns(df)
        logger.debug(f"Query returned {len(df)} rows")
        logger.debug(f"Columns in response: {df.columns.tolist()}")

        if hybrid_search_enabled_flag and not isinstance(db_handler, KeywordSearchBase):
            raise ValueError(f"Hybrid search is enabled but the db_handler {type(db_handler)} does not support it. ")
        # check if db_handler inherits from KeywordSearchBase
        if hybrid_search_enabled_flag and isinstance(db_handler, KeywordSearchBase):
            # If query_text is present, use it for keyword search
            logger.debug(f"Performing keyword search with query text: {query_text}")
            keyword_search_args = KeywordSearchArgs(query=query_text, column=TableField.CONTENT.value)
            keyword_query_obj = copy.deepcopy(query)

            keyword_query_obj.targets = [
                Identifier(TableField.ID.value),
                Identifier(TableField.CONTENT.value),
                Identifier(TableField.METADATA.value),
            ]

            df_keyword_select = db_handler.dispatch_select(
                keyword_query_obj, keyword_search_conditions, keyword_search_args=keyword_search_args
            )
            df_keyword_select = self.addapt_result_columns(df_keyword_select)
            logger.debug(f"Keyword search returned {len(df_keyword_select)} rows")
            logger.debug(f"Columns in keyword search response: {df_keyword_select.columns.tolist()}")
            # ensure df and df_keyword_select have exactly the same columns
            if not df_keyword_select.empty:
                if set(df.columns) != set(df_keyword_select.columns):
                    raise ValueError(
                        f"Keyword search returned different columns: {df_keyword_select.columns} "
                        f"than expected: {df.columns}"
                    )
                if hybrid_search_alpha:
                    df_keyword_select[TableField.DISTANCE.value] = (
                        hybrid_search_alpha * df_keyword_select[TableField.DISTANCE.value]
                    )
                    df[TableField.DISTANCE.value] = (1 - hybrid_search_alpha) * df[TableField.DISTANCE.value]
                df = pd.concat([df, df_keyword_select], ignore_index=True)
                # sort by distance if distance column exists
                if TableField.DISTANCE.value in df.columns:
                    df = df.sort_values(by=TableField.DISTANCE.value, ascending=True)
                # if chunk_id column exists remove duplicates based on chunk_id
                if "chunk_id" in df.columns:
                    df = df.drop_duplicates(subset=["chunk_id"])

        # Check if we have a rerank_model configured in KB params
        df = self.add_relevance(df, query_text, relevance_threshold, disable_reranking)

        return df

    def _get_allowed_metadata_columns(self) -> List[str] | None:
        # Return list of KB columns to restrict querying, if None: no restrictions

        if self._kb.params.get("version", 0) < 2:
            # disable for old version KBs
            return None

        user_columns = self._kb.params.get("metadata_columns", [])
        dynamic_columns = self._kb.params.get("inserted_metadata", [])

        columns = set(user_columns) | set(dynamic_columns)
        return [col.lower() for col in columns]

    def score_documents(self, query_text, documents, reranking_model_params):
        reranker = get_reranking_model_from_params(reranking_model_params)
        return reranker.get_scores(query_text, documents)

    def add_relevance(self, df, query_text, relevance_threshold=None, disable_reranking=False):
        relevance_column = TableField.RELEVANCE.value

        reranking_model_params = get_model_params(self._kb.params.get("reranking_model"), "default_reranking_model")
        if reranking_model_params and query_text and len(df) > 0 and not disable_reranking:
            # Use reranker for relevance score

            logger.info(f"Using knowledge reranking model from params: {reranking_model_params}")
            # Apply custom filtering threshold if provided
            if relevance_threshold is not None:
                reranking_model_params["filtering_threshold"] = relevance_threshold
                logger.info(f"Using custom filtering threshold: {relevance_threshold}")

            reranker = get_reranking_model_from_params(reranking_model_params)
            # Get documents to rerank
            documents = df["chunk_content"].tolist()
            # Use the get_scores method with disable_events=True
            scores = reranker.get_scores(query_text, documents)
            # Add scores as the relevance column
            df[relevance_column] = scores

            # Filter by threshold
            scores_array = np.array(scores)
            df = df[scores_array > reranker.filtering_threshold]
            logger.debug(f"Applied reranking with params: {reranking_model_params}")

        elif "distance" in df.columns:
            # Calculate relevance from distance
            logger.info("Calculating relevance from vector distance")
            df[relevance_column] = 1 / (1 + df["distance"])
            if relevance_threshold is not None:
                df = df[df[relevance_column] > relevance_threshold]

        else:
            df[relevance_column] = None
            df["distance"] = None
        # Sort by relevance
        df = df.sort_values(by=relevance_column, ascending=False)
        return df

    def addapt_conditions_columns(self, conditions):
        if conditions is None:
            return
        for condition in conditions:
            if condition.column in self.kb_to_vector_columns:
                condition.column = self.kb_to_vector_columns[condition.column]

    def addapt_result_columns(self, df):
        col_update = {}
        for kb_col, vec_col in self.kb_to_vector_columns.items():
            if vec_col in df.columns:
                col_update[vec_col] = kb_col

        df = df.rename(columns=col_update)

        columns = list(df.columns)
        # update id, get from metadata
        df[TableField.ID.value] = df[TableField.METADATA.value].apply(
            lambda m: None if m is None else m.get(self.kb_to_vector_columns["id"])
        )

        # id on first place
        return df[[TableField.ID.value] + columns]

    def insert_files(self, file_names: List[str]):
        """Process and insert files"""
        if not self.document_loader:
            raise ValueError("Document loader not configured")

        documents = list(self.document_loader.load_files(file_names))
        if documents:
            self.insert_documents(documents)

    def insert_web_pages(self, urls: List[str], crawl_depth: int, limit: int, filters: List[str] = None):
        """Process and insert web pages"""
        if not self.document_loader:
            raise ValueError("Document loader not configured")

        documents = list(
            self.document_loader.load_web_pages(urls, limit=limit, crawl_depth=crawl_depth, filters=filters)
        )
        if documents:
            self.insert_documents(documents)

    def insert_query_result(self, query: str, project_name: str):
        """Process and insert SQL query results"""
        ast_query = parse_sql(query)

        command_executor = ExecuteCommands(self.session)
        response = command_executor.execute_command(ast_query, project_name)

        if response.error_code is not None:
            raise ValueError(f"Error executing query: {response.error_message}")

        if response.data is None:
            raise ValueError("Query returned no data")

        records = response.data.records
        df = pd.DataFrame(records)

        self.insert(df)

    def insert_rows(self, rows: List[Dict]):
        """Process and insert raw data rows"""
        if not rows:
            return
        if len(rows) > MAX_INSERT_BATCH_SIZE:
            raise ValueError("Input data is too large, please load data in batches")

        df = pd.DataFrame(rows)

        self.insert(df)

    def insert_documents(self, documents: List[Document]):
        """Process and insert documents with preprocessing if configured"""
        df = pd.DataFrame([doc.model_dump() for doc in documents])

        self.insert(df)

    def update_query(self, query: Update):
        # add embeddings to content in updated collumns
        query = copy.deepcopy(query)

        emb_col = TableField.EMBEDDINGS.value
        cont_col = TableField.CONTENT.value

        db_handler = self.get_vector_db()
        conditions = db_handler.extract_conditions(query.where)
        doc_id = None
        for condition in conditions:
            if condition.column == "chunk_id" and condition.op == FilterOperator.EQUAL:
                doc_id = condition.value

        if cont_col in query.update_columns:
            content = query.update_columns[cont_col]

            # Apply preprocessing to content if configured
            if self.document_preprocessor:
                doc = Document(
                    id=doc_id,
                    content=content.value,
                    metadata={},  # Empty metadata for content-only updates
                )
                processed_chunks = self.document_preprocessor.process_documents([doc])
                if processed_chunks:
                    content.value = processed_chunks[0].content

            query.update_columns[emb_col] = Constant(self._content_to_embeddings(content))

        if "metadata" not in query.update_columns:
            query.update_columns["metadata"] = Constant({})

        # TODO search content in where clause?

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        self.addapt_conditions_columns(conditions)
        db_handler.dispatch_update(query, conditions)

    def delete_query(self, query: Delete):
        """
        Handles delete query to KB table.
        Replaces content values with embeddings in WHERE clause. Sends query to vector db
        :param query: query to KB table
        """
        query_traversal(query.where, self._replace_query_content)

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        db_handler = self.get_vector_db()
        conditions = db_handler.extract_conditions(query.where)
        self.addapt_conditions_columns(conditions)
        db_handler.dispatch_delete(query, conditions)

    def hybrid_search(
        self,
        query: str,
        keywords: List[str] = None,
        metadata: Dict[str, str] = None,
        distance_function=DistanceFunction.COSINE_DISTANCE,
    ) -> pd.DataFrame:
        query_df = pd.DataFrame.from_records([{TableField.CONTENT.value: query}])
        embeddings_df = self._df_to_embeddings(query_df)
        if embeddings_df.empty:
            return pd.DataFrame([])
        embeddings = embeddings_df.iloc[0][TableField.EMBEDDINGS.value]
        keywords_query = None
        if keywords is not None:
            keywords_query = " ".join(keywords)
        db_handler = self.get_vector_db()
        return db_handler.hybrid_search(
            self._kb.vector_database_table,
            embeddings,
            query=keywords_query,
            metadata=metadata,
            distance_function=distance_function,
        )

    def clear(self):
        """
        Clear data in KB table
        Sends delete to vector db table
        """
        db_handler = self.get_vector_db()
        db_handler.delete(self._kb.vector_database_table)

    def insert(self, df: pd.DataFrame, params: dict = None):
        """Insert dataframe to KB table.

        Args:
            df: DataFrame to insert
            params: User parameters of insert
        """
        if df.empty:
            return

        try:
            run_query_id = ctx.run_query_id
            # Link current KB to running query (where KB is used to insert data)
            if run_query_id is not None:
                self._kb.query_id = run_query_id
                db.session.commit()

        except AttributeError:
            ...

        # First adapt column names to identify content and metadata columns
        adapted_df = self._adapt_column_names(df)
        content_columns = self._kb.params.get("content_columns", [TableField.CONTENT.value])

        # Convert DataFrame rows to documents, creating separate documents for each content column
        raw_documents = []
        for idx, row in adapted_df.iterrows():
            base_metadata = self._parse_metadata(row.get(TableField.METADATA.value, {}))
            provided_id = row.get(TableField.ID.value)

            for col in content_columns:
                content = row.get(col)
                if content and str(content).strip():
                    content_str = str(content)

                    # Use provided_id directly if it exists, otherwise generate one
                    doc_id = self._generate_document_id(content_str, col, provided_id)

                    metadata = {
                        **base_metadata,
                        "_original_row_index": str(idx),  # provide link to original row index
                        "_content_column": col,
                    }

                    raw_documents.append(Document(content=content_str, id=doc_id, metadata=metadata))

        # Apply preprocessing to all documents if preprocessor exists
        if self.document_preprocessor:
            processed_chunks = self.document_preprocessor.process_documents(raw_documents)
        else:
            processed_chunks = raw_documents  # Use raw documents if no preprocessing

        # Convert processed chunks back to DataFrame with standard structure
        df = pd.DataFrame(
            [
                {
                    TableField.CONTENT.value: chunk.content,
                    TableField.ID.value: chunk.id,
                    TableField.METADATA.value: chunk.metadata,
                }
                for chunk in processed_chunks
            ]
        )

        if df.empty:
            logger.warning("No valid content found in any content columns")
            return

        # add embeddings and send to vector db
        df_emb = self._df_to_embeddings(df)
        df = pd.concat([df, df_emb], axis=1)
        db_handler = self.get_vector_db()

        if params is not None and params.get("kb_no_upsert", False):
            # speed up inserting by disable checking existing records
            db_handler.insert(self._kb.vector_database_table, df)
        else:
            db_handler.do_upsert(self._kb.vector_database_table, df)

    def _adapt_column_names(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert input columns for vector db input
        - id, content and metadata
        """
        # Debug incoming data
        logger.debug(f"Input DataFrame columns: {df.columns}")
        logger.debug(f"Input DataFrame first row: {df.iloc[0].to_dict()}")

        params = self._kb.params
        columns = list(df.columns)

        # -- prepare id --
        id_column = params.get("id_column")
        if id_column is not None and id_column not in columns:
            id_column = None

        if id_column is None and TableField.ID.value in columns:
            id_column = TableField.ID.value

        # Also check for case-insensitive 'id' column
        if id_column is None:
            column_map = {col.lower(): col for col in columns}
            if "id" in column_map:
                id_column = column_map["id"]

        if id_column is not None:
            columns.remove(id_column)
            logger.debug(f"Using ID column: {id_column}")

        # Create output dataframe
        df_out = pd.DataFrame()

        # Add ID if present
        if id_column is not None:
            df_out[TableField.ID.value] = df[id_column]
            logger.debug(f"Added IDs: {df_out[TableField.ID.value].tolist()}")

        # -- prepare content and metadata --
        content_columns = params.get("content_columns", [TableField.CONTENT.value])
        metadata_columns = params.get("metadata_columns")

        logger.debug(f"Processing with: content_columns={content_columns}, metadata_columns={metadata_columns}")

        # Handle SQL query result columns
        if content_columns:
            # Ensure content columns are case-insensitive
            column_map = {col.lower(): col for col in columns}
            content_columns = [column_map.get(col.lower(), col) for col in content_columns]
            logger.debug(f"Mapped content columns: {content_columns}")

        if metadata_columns:
            # Ensure metadata columns are case-insensitive
            column_map = {col.lower(): col for col in columns}
            metadata_columns = [column_map.get(col.lower(), col) for col in metadata_columns]
            logger.debug(f"Mapped metadata columns: {metadata_columns}")

        content_columns = list(set(content_columns).intersection(columns))
        if len(content_columns) == 0:
            raise ValueError(f"Content columns {params.get('content_columns')} not found in dataset: {columns}")

        if metadata_columns is not None:
            metadata_columns = list(set(metadata_columns).intersection(columns))
        else:
            # all the rest columns
            metadata_columns = list(set(columns).difference(content_columns))

            # update list of used columns
            inserted_metadata = set(self._kb.params.get("inserted_metadata", []))
            inserted_metadata.update(metadata_columns)
            self._kb.params["inserted_metadata"] = list(inserted_metadata)
            flag_modified(self._kb, "params")
            db.session.commit()

        # Add content columns directly (don't combine them)
        for col in content_columns:
            df_out[col] = df[col]

        # Add metadata
        if metadata_columns and len(metadata_columns) > 0:

            def convert_row_to_metadata(row):
                metadata = {}
                for col in metadata_columns:
                    value = row[col]
                    value_type = type(value)
                    # Convert numpy/pandas types to Python native types
                    if safe_pandas_is_datetime(value) or isinstance(value, pd.Timestamp):
                        value = str(value)
                    elif pd.api.types.is_integer_dtype(value_type):
                        value = int(value)
                    elif pd.api.types.is_float_dtype(value_type) or isinstance(value, decimal.Decimal):
                        value = float(value)
                    elif pd.api.types.is_bool_dtype(value_type):
                        value = bool(value)
                    elif isinstance(value, dict):
                        metadata.update(value)
                        continue
                    elif value is not None:
                        value = str(value)
                    metadata[col] = value
                return metadata

            metadata_dict = df[metadata_columns].apply(convert_row_to_metadata, axis=1)
            df_out[TableField.METADATA.value] = metadata_dict

        logger.debug(f"Output DataFrame columns: {df_out.columns}")
        logger.debug(f"Output DataFrame first row: {df_out.iloc[0].to_dict() if not df_out.empty else 'Empty'}")

        return df_out

    def _replace_query_content(self, node, **kwargs):
        if isinstance(node, BinaryOperation):
            if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Constant):
                col_name = node.args[0].parts[-1]
                if col_name.lower() == TableField.CONTENT.value:
                    # replace
                    node.args[0].parts = [TableField.EMBEDDINGS.value]
                    node.args[1].value = [self._content_to_embeddings(node.args[1].value)]

    def get_vector_db(self) -> VectorStoreHandler:
        """
        helper to get vector db handler
        """
        if self._vector_db is None:
            database = db.Integration.query.get(self._kb.vector_database_id)
            if database is None:
                raise ValueError("Vector database not found. Is it deleted?")
            database_name = database.name
            self._vector_db = self.session.integration_controller.get_data_handler(database_name)
        return self._vector_db

    def get_vector_db_table_name(self) -> str:
        """
        helper to get underlying table name used for embeddings
        """
        return self._kb.vector_database_table

    def _df_to_embeddings(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Returns embeddings for input dataframe.
        Uses model embedding model to convert content to embeddings.
        Automatically detects input and output of model using model description
        :param df:
        :return: dataframe with embeddings
        """

        if df.empty:
            return pd.DataFrame([], columns=[TableField.EMBEDDINGS.value])

        model_id = self._kb.embedding_model_id

        if model_id is None:
            # call litellm handler
            messages = list(df[TableField.CONTENT.value])
            embedding_params = get_model_params(self._kb.params.get("embedding_model", {}), "default_embedding_model")
            results = self.call_litellm_embedding(self.session, embedding_params, messages)
            results = [[val] for val in results]
            return pd.DataFrame(results, columns=[TableField.EMBEDDINGS.value])

        # get the input columns
        model_rec = db.session.query(db.Predictor).filter_by(id=model_id).first()

        assert model_rec is not None, f"Model not found: {model_id}"
        model_project = db.session.query(db.Project).filter_by(id=model_rec.project_id).first()

        project_datanode = self.session.datahub.get(model_project.name)

        model_using = model_rec.learn_args.get("using", {})
        input_col = model_using.get("question_column")
        if input_col is None:
            input_col = model_using.get("input_column")

        if input_col is not None and input_col != TableField.CONTENT.value:
            df = df.rename(columns={TableField.CONTENT.value: input_col})

        df_out = project_datanode.predict(model_name=model_rec.name, df=df, params=self.model_params)

        target = model_rec.to_predict[0]
        if target != TableField.EMBEDDINGS.value:
            # adapt output for vectordb
            df_out = df_out.rename(columns={target: TableField.EMBEDDINGS.value})

        df_out = df_out[[TableField.EMBEDDINGS.value]]

        return df_out

    def _content_to_embeddings(self, content: str) -> List[float]:
        """
        Converts string to embeddings
        :param content: input string
        :return: embeddings
        """
        df = pd.DataFrame([[content]], columns=[TableField.CONTENT.value])
        res = self._df_to_embeddings(df)
        return res[TableField.EMBEDDINGS.value][0]

    @staticmethod
    def call_litellm_embedding(session, model_params, messages):
        args = copy.deepcopy(model_params)

        if "model_name" not in args:
            raise ValueError("'model_name' must be provided for embedding model")

        llm_model = args.pop("model_name")
        engine = args.pop("provider")

        module = session.integration_controller.get_handler_module("litellm")
        if module is None or module.Handler is None:
            raise ValueError(f'Unable to use "{engine}" provider. Litellm handler is not installed')
        return module.Handler.embeddings(engine, llm_model, messages, args)

    def build_rag_pipeline(self, retrieval_config: dict):
        """
        Builds a RAG pipeline with returned sources

        Args:
            retrieval_config: dict with retrieval config

        Returns:
            RAG: Configured RAG pipeline instance

        Raises:
            ValueError: If the configuration is invalid or required components are missing
        """
        # Get embedding model from knowledge base
        embeddings_model = None
        embedding_model_params = get_model_params(self._kb.params.get("embedding_model", {}), "default_embedding_model")
        if self._kb.embedding_model:
            # Extract embedding model args from knowledge base table
            embedding_args = self._kb.embedding_model.learn_args.get("using", {})
            # Construct the embedding model directly
            embeddings_model = construct_model_from_args(embedding_args)
            logger.debug(f"Using knowledge base embedding model with args: {embedding_args}")
        elif embedding_model_params:
            embeddings_model = get_embedding_model_from_params(embedding_model_params)
            logger.debug(f"Using knowledge base embedding model from params: {self._kb.params['embedding_model']}")
        else:
            embeddings_model = DEFAULT_EMBEDDINGS_MODEL_CLASS()
            logger.debug("Using default embedding model as knowledge base has no embedding model")

        # Update retrieval config with knowledge base parameters
        kb_params = {"vector_store_config": {"kb_table": self}}

        # Load and validate config
        try:
            rag_config = load_rag_config(retrieval_config, kb_params, embeddings_model)

            # Build LLM if specified
            if "llm_model_name" in rag_config:
                llm_args = {"model_name": rag_config.llm_model_name}
                if not rag_config.llm_provider:
                    llm_args["provider"] = get_llm_provider(llm_args)
                else:
                    llm_args["provider"] = rag_config.llm_provider
                rag_config.llm = create_chat_model(llm_args)

            # Create RAG pipeline
            rag = RAG(rag_config)
            logger.debug(f"RAG pipeline created with config: {rag_config}")
            return rag

        except Exception as e:
            logger.error(f"Error building RAG pipeline: {str(e)}")
            raise ValueError(f"Failed to build RAG pipeline: {str(e)}")

    def _parse_metadata(self, base_metadata):
        """Helper function to robustly parse metadata string to dict"""
        if isinstance(base_metadata, dict):
            return base_metadata
        if isinstance(base_metadata, str):
            try:
                import ast

                return ast.literal_eval(base_metadata)
            except (SyntaxError, ValueError):
                logger.warning(f"Could not parse metadata: {base_metadata}. Using empty dict.")
                return {}
        return {}

    def _generate_document_id(self, content: str, content_column: str, provided_id: str = None) -> str:
        """Generate a deterministic document ID using the utility function."""
        from mindsdb.interfaces.knowledge_base.utils import generate_document_id

        return generate_document_id(content=content, provided_id=provided_id)

    def _convert_metadata_value(self, value):
        """
        Convert metadata value to appropriate Python type.

        Args:
            value: The value to convert

        Returns:
            Converted value in appropriate Python type
        """
        if pd.isna(value):
            return None

        # Handle pandas/numpy types
        if pd.api.types.is_datetime64_any_dtype(value) or isinstance(value, pd.Timestamp):
            return str(value)
        elif pd.api.types.is_integer_dtype(type(value)):
            return int(value)
        elif pd.api.types.is_float_dtype(type(value)):
            return float(value)
        elif pd.api.types.is_bool_dtype(type(value)):
            return bool(value)

        # Handle basic Python types
        if isinstance(value, (int, float, bool)):
            return value

        # Convert everything else to string
        return str(value)

    def create_index(self):
        """
        Create an index on the knowledge base table
        :param index_name: name of the index
        :param params: parameters for the index
        """
        db_handler = self.get_vector_db()
        db_handler.create_index(self._kb.vector_database_table)


class KnowledgeBaseController:
    """
    Knowledge base controller handles all
    manages knowledge bases
    """

    KB_VERSION = 2

    def __init__(self, session) -> None:
        self.session = session

    def add(
        self,
        name: str,
        project_name: str,
        storage: Identifier,
        params: dict,
        preprocessing_config: Optional[dict] = None,
        if_not_exists: bool = False,
        keyword_search_enabled: bool = False,
        # embedding_model: Identifier = None, # Legacy: Allow MindsDB models to be passed as embedding_model.
    ) -> db.KnowledgeBase:
        """
        Add a new knowledge base to the database
        :param preprocessing_config: Optional preprocessing configuration to validate and store
        :param is_sparse: Whether to use sparse vectors for embeddings
        :param vector_size: Optional size specification for vectors, required when is_sparse=True
        """
        if not name.islower():
            raise ValueError(f"The name must be in lower case: {name}")

        # fill variables
        params = variables_controller.fill_parameters(params)

        try:
            KnowledgeBaseInputParams.model_validate(params)
        except ValidationError as e:
            problems = []
            for error in e.errors():
                parameter = ".".join([str(i) for i in error["loc"]])
                param_type = error["type"]
                if param_type == "extra_forbidden":
                    msg = f"Parameter '{parameter}' is not allowed"
                else:
                    msg = f"Error in '{parameter}' (type: {param_type}): {error['msg']}. Input: {repr(error['input'])}"
                problems.append(msg)

            msg = "\n".join(problems)
            if len(problems) > 1:
                msg = "\n" + msg
            raise ValueError(f"Problem with knowledge base parameters: {msg}")

        # Validate preprocessing config first if provided
        if preprocessing_config is not None:
            PreprocessingConfig(**preprocessing_config)  # Validate before storing
            params = params or {}
            params["preprocessing"] = preprocessing_config

        # Check if vector_size is provided when using sparse vectors
        is_sparse = params.get("is_sparse")
        vector_size = params.get("vector_size")
        if is_sparse and vector_size is None:
            raise ValueError("vector_size is required when is_sparse=True")

        # get project id
        project = self.session.database_controller.get_project(project_name)
        project_id = project.id

        # not difference between cases in sql
        name = name.lower()
        # check if knowledge base already exists
        kb = self.get(name, project_id)
        if kb is not None:
            if if_not_exists:
                return kb
            raise EntityExistsError("Knowledge base already exists", name)

        embedding_params = get_model_params(params.get("embedding_model", {}), "default_embedding_model")

        # if model_name is None:  # Legacy
        model_name = self._create_embedding_model(
            project.name,
            params=embedding_params,
            kb_name=name,
        )
        if model_name is not None:
            params["created_embedding_model"] = model_name

        embedding_model_id = None
        if model_name is not None:
            model = self.session.model_controller.get_model(name=model_name, project_name=project.name)
            model_record = db.Predictor.query.get(model["id"])
            embedding_model_id = model_record.id

        # if params.get("reranking_model", {}) is bool and False we evaluate it to empty dictionary
        reranking_model_params = params.get("reranking_model", {})

        if isinstance(reranking_model_params, bool) and not reranking_model_params:
            params["reranking_model"] = {}
        # if params.get("reranking_model", {}) is string and false in any case we evaluate it to empty dictionary
        if isinstance(reranking_model_params, str) and reranking_model_params.lower() == "false":
            params["reranking_model"] = {}

        reranking_model_params = get_model_params(reranking_model_params, "default_reranking_model")
        if reranking_model_params:
            # Get reranking model from params.
            # This is called here to check validaity of the parameters.
            try:
                reranker = get_reranking_model_from_params(reranking_model_params)
                reranker.get_scores("test", ["test"])
            except (ValueError, RuntimeError) as e:
                raise RuntimeError(f"Problem with reranker config: {e}")

        # search for the vector database table
        if storage is None:
            cloud_pg_vector = os.environ.get("KB_PGVECTOR_URL")
            if cloud_pg_vector:
                vector_table_name = name
                # Add sparse vector support for pgvector
                vector_db_params = {}
                # Check both explicit parameter and model configuration
                is_sparse = is_sparse or model_record.learn_args.get("using", {}).get("sparse")
                if is_sparse:
                    vector_db_params["is_sparse"] = True
                    if vector_size is not None:
                        vector_db_params["vector_size"] = vector_size
                vector_db_name = self._create_persistent_pgvector(vector_db_params)

            else:
                # create chroma db with same name
                vector_table_name = "default_collection"
                vector_db_name = self._create_persistent_chroma(name)
                # memorize to remove it later
                params["default_vector_storage"] = vector_db_name
        elif len(storage.parts) != 2:
            raise ValueError("Storage param has to be vector db with table")
        else:
            vector_db_name, vector_table_name = storage.parts

        data_node = self.session.datahub.get(vector_db_name)
        if data_node:
            vector_store_handler = data_node.integration_handler
        else:
            raise ValueError(
                f"Unable to find database named {vector_db_name}, please make sure {vector_db_name} is defined"
            )
        # create table in vectordb before creating KB
        vector_store_handler.create_table(vector_table_name)
        if keyword_search_enabled:
            vector_store_handler.add_full_text_index(vector_table_name, TableField.CONTENT.value)
        vector_database_id = self.session.integration_controller.get(vector_db_name)["id"]

        # Store sparse vector settings in params if specified
        if is_sparse:
            params = params or {}
            params["vector_config"] = {"is_sparse": is_sparse}
            if vector_size is not None:
                params["vector_config"]["vector_size"] = vector_size

        params["version"] = self.KB_VERSION
        kb = db.KnowledgeBase(
            name=name,
            project_id=project_id,
            vector_database_id=vector_database_id,
            vector_database_table=vector_table_name,
            embedding_model_id=embedding_model_id,
            params=params,
        )
        db.session.add(kb)
        db.session.commit()
        return kb

    def _create_persistent_pgvector(self, params=None):
        """Create default vector database for knowledge base, if not specified"""
        vector_store_name = "kb_pgvector_store"

        # check if exists
        if self.session.integration_controller.get(vector_store_name):
            return vector_store_name

        self.session.integration_controller.add(vector_store_name, "pgvector", params or {})
        return vector_store_name

    def _create_persistent_chroma(self, kb_name, engine="chromadb"):
        """Create default vector database for knowledge base, if not specified"""

        vector_store_name = f"{kb_name}_{engine}"

        vector_store_folder_name = f"{vector_store_name}"
        connection_args = {"persist_directory": vector_store_folder_name}

        # check if exists
        if self.session.integration_controller.get(vector_store_name):
            return vector_store_name

        self.session.integration_controller.add(vector_store_name, engine, connection_args)
        return vector_store_name

    def _create_embedding_model(self, project_name, engine="openai", params: dict = None, kb_name=""):
        """create a default embedding model for knowledge base, if not specified"""
        model_name = f"kb_embedding_{kb_name}"

        # drop if exists - parameters can be different
        try:
            model = self.session.model_controller.get_model(model_name, project_name=project_name)
            if model is not None:
                self.session.model_controller.delete_model(model_name, project_name)
        except PredictorRecordNotFound:
            pass

        if "provider" not in params:
            raise ValueError("'provider' parameter is required for embedding model")

        # check available providers
        avail_providers = ("openai", "azure_openai", "bedrock", "gemini", "google")
        if params["provider"] not in avail_providers:
            raise ValueError(
                f"Wrong embedding provider: {params['provider']}. Available providers: {', '.join(avail_providers)}"
            )

        if params["provider"] not in ("openai", "azure_openai"):
            # try use litellm
            try:
                KnowledgeBaseTable.call_litellm_embedding(self.session, params, ["test"])
            except Exception as e:
                raise RuntimeError(f"Problem with embedding model config: {e}")
            return

        if "provider" in params:
            engine = params.pop("provider").lower()

        api_key = get_api_key(engine, params, strict=False)
        if api_key is None:
            if "api_key" in params:
                params.pop("api_key")
            else:
                raise ValueError("'api_key' parameter is required for embedding model")

        if engine == "azure_openai":
            engine = "openai"
            params["provider"] = "azure"

        if engine == "openai":
            if "question_column" not in params:
                params["question_column"] = "content"
            if api_key:
                params[f"{engine}_api_key"] = api_key
                if "api_key" in params:
                    params.pop("api_key")
            if "base_url" in params:
                params["api_base"] = params.pop("base_url")

        params["engine"] = engine
        params["join_learn_process"] = True
        params["mode"] = "embedding"

        # Include API key if provided.
        statement = CreatePredictor(
            name=Identifier(parts=[project_name, model_name]),
            using=params,
            targets=[Identifier(parts=[TableField.EMBEDDINGS.value])],
        )

        command_executor = ExecuteCommands(self.session)
        resp = command_executor.answer_create_predictor(statement, project_name)
        # check model status
        record = resp.data.records[0]
        if record["STATUS"] == "error":
            raise ValueError("Embedding model error:" + record["ERROR"])
        return model_name

    def delete(self, name: str, project_name: int, if_exists: bool = False) -> None:
        """
        Delete a knowledge base from the database
        """
        try:
            project = self.session.database_controller.get_project(project_name)
        except ValueError:
            raise ValueError(f"Project not found: {project_name}")
        project_id = project.id

        # check if knowledge base exists
        kb = self.get(name, project_id)
        if kb is None:
            # knowledge base does not exist
            if if_exists:
                return
            else:
                raise EntityNotExistsError("Knowledge base does not exist", name)

        # kb exists
        db.session.delete(kb)
        db.session.commit()

        # drop objects if they were created automatically
        if "default_vector_storage" in kb.params:
            try:
                handler = self.session.datahub.get(kb.params["default_vector_storage"]).integration_handler
                handler.drop_table(kb.vector_database_table)
                self.session.integration_controller.delete(kb.params["default_vector_storage"])
            except EntityNotExistsError:
                pass
        if "created_embedding_model" in kb.params:
            try:
                self.session.model_controller.delete_model(kb.params["created_embedding_model"], project_name)
            except EntityNotExistsError:
                pass

    def get(self, name: str, project_id: int) -> db.KnowledgeBase:
        """
        Get a knowledge base from the database
        by name + project_id
        """
        kb = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                name=name,
                project_id=project_id,
            )
            .first()
        )
        return kb

    def get_table(self, name: str, project_id: int, params: dict = None) -> KnowledgeBaseTable:
        """
        Returns kb table object with properly configured preprocessing
        :param name: table name
        :param project_id: project id
        :param params: runtime parameters for KB. Keys: 'model' - parameters for embedding model
        :return: kb table object
        """
        kb = self.get(name, project_id)
        if kb is not None:
            table = KnowledgeBaseTable(kb, self.session)
            if params:
                table.model_params = params.get("model")

            # Always configure preprocessing - either from params or default
            if kb.params and "preprocessing" in kb.params:
                table.configure_preprocessing(kb.params["preprocessing"])
            else:
                table.configure_preprocessing(None)  # This ensures default preprocessor is created

            return table

    def list(self, project_name: str = None) -> List[dict]:
        """
        List all knowledge bases from the database
        belonging to a project
        """
        project_controller = ProjectController()
        projects = project_controller.get_list()
        if project_name is not None:
            projects = [p for p in projects if p.name == project_name]

        query = db.session.query(db.KnowledgeBase).filter(
            db.KnowledgeBase.project_id.in_(list([p.id for p in projects]))
        )

        data = []
        project_names = {i.id: i.name for i in project_controller.get_list()}

        for record in query:
            kb = record.as_dict(with_secrets=self.session.show_secrets)
            kb["project_name"] = project_names[record.project_id]

            data.append(kb)

        return data

    def create_index(self, table_name, project_name):
        project_id = self.session.database_controller.get_project(project_name).id
        kb_table = self.get_table(table_name, project_id)
        kb_table.create_index()

    def update(self, name: str, project_id: int, **kwargs) -> db.KnowledgeBase:
        """
        Update a knowledge base record
        """
        raise NotImplementedError()

    def evaluate(self, table_name: str, project_name: str, params: dict = None) -> pd.DataFrame:
        """
        Run evaluate and/or create test data for evaluation
        :param table_name: name of KB
        :param project_name: project of KB
        :param params: evaluation parameters
        :return: evaluation results
        """
        project_id = self.session.database_controller.get_project(project_name).id
        kb_table = self.get_table(table_name, project_id)

        scores = EvaluateBase.run(self.session, kb_table, params)

        return scores



import base64
import datetime
import os
import json
import select
import socketserver
import struct
import sys
from functools import partial
import socket
from typing import Callable, Dict, Type, Any, Iterable, Sequence

from mindsdb.api.executor.controllers import SessionController
from mindsdb.api.postgres.postgres_proxy.executor import Executor
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import CHARSET_NUMBERS
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.api.common.middleware import check_auth
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import SQLAnswer
from mindsdb.api.postgres.postgres_proxy.postgres_packets.errors import POSTGRES_SYNTAX_ERROR_CODE
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_fields import GenericField, PostgresField
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message_formats import (
    Terminate,
    Query,
    AuthenticationClearTextPassword,
    AuthenticationOk,
    RowDescriptions,
    DataRow,
    CommandComplete,
    ReadyForQuery,
    ConnectionFailure,
    ParameterStatus,
    Error,
    Execute,
    Bind,
    Parse,
    Sync,
    ParseComplete,
    InvalidSQLStatementName,
    BindComplete,
    Describe,
    DataException,
    ParameterDescription,
)
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message import PostgresMessage
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_packets import (
    PostgresPacketReader,
    PostgresPacketBuilder,
)
from mindsdb.api.postgres.postgres_proxy.utilities import strip_null_byte
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
from mindsdb.api.mysql.mysql_proxy.external_libs.mysql_scramble import scramble as scramble_func


class PostgresProxyHandler(socketserver.StreamRequestHandler):
    client_buffer: PostgresPacketReader
    user_parameters: Dict[bytes, bytes]

    def __init__(self, request, client_address, server):
        self.logger = log.getLogger(__name__)
        self.charset = "utf8"
        self.charset_text_type = CHARSET_NUMBERS["utf8_general_ci"]
        self.session = None
        self.client_capabilities = None
        self.user_parameters = None
        self.named_statements = {}
        self.unnamed_statement = None
        self.is_cloud = False
        self.named_portals = {}
        self.unnamed_portal = None
        self.transaction_status = b"I"  # I: Idle, T: Transaction Block, E: Failed Transaction Block
        super().__init__(request, client_address, server)

    def handle(self) -> None:
        ctx.set_default()
        self.init_session()
        self.logger.debug("handle new incoming connection")
        cloud_connection = self.is_cloud_connection()
        if cloud_connection["is_cloud"]:
            ctx.company_id = cloud_connection.get("company_id")
            self.is_cloud = True

        self.message_map: Dict[Type[PostgresMessage], Callable[[Any], bool]] = {
            Terminate: self.terminate,
            Query: self.query,
            Parse: self.parse,
            Bind: self.bind,
            Execute: self.execute,
            Describe: self.describe,
            Sync: self.sync,
        }
        self.client_buffer = PostgresPacketReader(self.rfile)
        if self.is_cloud:
            # We already have a connection started through the gateway.
            started = True
            self.handshake()
        else:
            started = self.start_connection()
        if started:
            self.logger.debug("connection started")
            self.send_initial_data()
            self.main_loop()

    def is_cloud_connection(self):
        """Determine source of connection. Must be call before handshake.
        Idea based on: real mysql connection does not send anything before server handshake, so
        socket should be in 'out' state. In opposite, clout connection sends '0000' right after
        connection. '0000' selected because in real mysql connection it should be length of package,
        and it can not be 0.

        Copied from mysql_proxy
        TODO: Extract method into common
        """
        is_cloud = config.get("cloud", False)

        if sys.platform != "linux" or is_cloud is False:
            return {"is_cloud": False}

        read_poller = select.poll()
        read_poller.register(self.request, select.POLLIN)
        events = read_poller.poll(30)

        if len(events) == 0:
            return {"is_cloud": False}

        first_byte = self.request.recv(4, socket.MSG_PEEK)
        if first_byte == b"\x00\x00\x00\x00":
            self.request.recv(4)
            client_capabilities = self.request.recv(8)
            client_capabilities = struct.unpack("L", client_capabilities)[0]

            company_id = self.request.recv(4)
            company_id = struct.unpack("I", company_id)[0]

            user_class = self.request.recv(1)
            user_class = struct.unpack("B", user_class)[0]

            database_name_len = self.request.recv(2)
            database_name_len = struct.unpack("H", database_name_len)[0]

            database_name = ""
            if database_name_len > 0:
                database_name = self.request.recv(database_name_len).decode()

            return {
                "is_cloud": True,
                "client_capabilities": client_capabilities,
                "company_id": company_id,
                "user_class": user_class,
                "database": database_name,
            }

        return {"is_cloud": False}

    def parse(self, message: Parse):
        self.logger.info("Postgres_Proxy: Parsing")
        executor = Executor(session=self.session, proxy_server=self, charset=self.charset)
        # TODO: Remove comment if unneeded ot use session since we're storing in this proxy class per session anyway
        # stmt_id = self.session.register_stmt(executor)
        executor.stmt_prepare(sql=message.query)
        statement = {"executor": executor, "parse": message}
        if message.name:
            self.named_statements[message.name] = statement
        else:
            self.unnamed_statement = statement
        self.send(ParseComplete())
        return True

    def bind(self, message: Bind):
        self.logger.info(f"Postgres_Proxy: Binding {message.name} with params {message.parameters}")
        if message.statement_name:
            statement = self.named_statements[message.statement_name]
        elif self.unnamed_statement:
            statement = self.unnamed_statement
        else:
            self.send(InvalidSQLStatementName())
            return True

        # TODO Should check validity of statement here and not at parse stage
        portal = statement.copy()
        portal["bind"] = message
        if message.name:
            self.named_portals[message.name] = portal
        else:
            self.unnamed_portal = portal
        self.send(BindComplete())
        return True

    def describe(self, message: Describe):
        self.logger.info("Postgres_Proxy: Describing")
        if message.describe_type == b"P":
            if message.name:
                describing = self.named_portals[message.name]
            elif self.unnamed_statement:
                describing = self.unnamed_portal
            else:
                self.send(InvalidSQLStatementName("Portal Does not Exist"))
                return True
        elif message.describe_type == b"S":
            if message.name:
                describing = self.named_statements[message.name]
            elif self.unnamed_statement:
                describing = self.unnamed_statement
            else:
                self.send(InvalidSQLStatementName())
                return True
            self.send(ParameterDescription(parameters=describing["parse"]["parameters"]))
        else:
            self.send(DataException(message="Describe did not have correct type. Can be 'P' or 'S'"))
            return True

        fields = self.to_postgres_fields(describing["executor"].columns)
        self.send(RowDescriptions(fields=fields))
        return True

    def execute(self, message: Execute):
        self.logger.info("Postgres_Proxy: Executing")
        if message.name:
            portal = self.named_portals[message.name]
        elif self.unnamed_portal:
            portal = self.unnamed_portal
        else:
            self.send(InvalidSQLStatementName("Portal does not exist"))

        executor = portal["executor"]
        params = portal["bind"].parameters
        executor.stmt_execute(param_values=params)
        sql_answer = self.return_executor_data(executor)
        self.respond_from_sql_answer(sql=executor.sql, sql_answer=sql_answer, row_descs=False)
        return True

    def sync(self, message: Sync):
        self.logger.info("Postgres_Proxy: Syncing")
        # TODO: Close/commit transaction if outside of a block. Maybe no collaries since Proxy
        self.send_ready()
        return True

    def init_session(self):
        self.logger.info("New connection [{ip}:{port}]".format(ip=self.client_address[0], port=self.client_address[1]))
        self.logger.debug(self.__dict__)

        if self.server.connection_id >= 65025:
            self.server.connection_id = 0
        self.server.connection_id += 1
        self.connection_id = self.server.connection_id
        self.session = SessionController()
        self.session.database = config.get("default_project")

        if hasattr(self.server, "salt") and isinstance(self.server.salt, str):
            self.salt = self.server.salt
        else:
            self.salt = base64.b64encode(os.urandom(15)).decode()

        self.socket = self.request

        self.current_transaction = None

        self.logger.debug("session salt: {salt}".format(salt=self.salt))

    def process_query(self, sql):
        executor = Executor(session=self.session, proxy_server=self, charset=self.charset)
        self.logger.debug("processing query\n%s", sql)
        try:
            executor.query_execute(sql)
        except Exception as e:
            return SQLAnswer(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=str(e).encode(self.get_encoding()),
                error_code=POSTGRES_SYNTAX_ERROR_CODE.encode(self.get_encoding()),
            )
        return self.return_executor_data(executor)

    def return_executor_data(self, executor):
        if executor.data is None:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.OK,
                state_track=executor.state_track,
            )
        else:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.TABLE,
                state_track=executor.state_track,
                result_set=executor.data,
                status=executor.server_status,
            )
        return resp

    def start_connection(self):
        self.logger.debug("starting handshake")
        self.handshake()
        self.logger.debug("handshake complete, checking authentication")
        return self.authenticate()

    def send(self, message: PostgresMessage):
        self.logger.debug("Sending message of type %s" % message.__class__.__name__)
        message.send(self.wfile)

    def handshake(self):
        self.client_buffer.read_verify_ssl_request()
        # self.send(NoticeResponse()) -- Should Probably not send. Looks in protocol manual to be sent for warning
        self.logger.debug("Sending No to SSL Request")
        PostgresPacketBuilder().write_char(b"N", self.wfile)
        self.user_parameters = self.client_buffer.read_startup_message()

    def authenticate(self, ask_for_password=False):
        if ask_for_password:
            self.send(AuthenticationClearTextPassword())
            password = self.client_buffer.read_authentication(encoding=self.charset)
        else:
            password = ""
        username = self.user_parameters[b"user"].decode(encoding=self.charset)
        auth_data = self.server.check_auth(username, password, scramble_func, self.salt, ctx.company_id)
        if auth_data["success"]:
            self.logger.debug("Authentication succeeded")
            self.session.username = auth_data["username"]
            self.session.auth = True
            self.send(AuthenticationOk())
            return True
        else:
            if not ask_for_password:  # try asking for password
                return self.authenticate(ask_for_password=True)
            self.logger.debug("Authentication failed")
            self.send(ConnectionFailure(message="Authentication failed."))
            return False

    def terminate(self, message: Terminate) -> bool:
        self.logger.info("Postgres_Proxy: Terminating")
        return False

    def get_encoding(self) -> str:
        return self.user_parameters.get(b"user_encoding", None) or self.charset

    def return_ok(self, sql, rows: int = 0):
        command = self.get_command(sql)
        if command == "BEGIN":
            self.transaction_status = b"T"
        if command == "COMMIT":
            self.transaction_status = b"I"
        if command == "CREATE":
            command = "SELECT"
        if command in ("INSERT", "DELETE", "UPDATE", "SELECT", "MOVE", "FETCH", "COPY"):
            command = f"{command} {rows}"
        self.send(CommandComplete(tag=command.encode(encoding=self.get_encoding())))
        return True

    def get_command(self, sql):
        sql = self.stringify_sql(sql)
        return sql.split(" ", 1)[0]

    def stringify_sql(self, sql):
        if type(sql) == bytes:
            encoding = self.get_encoding()
            sql: str = sql.decode(encoding)
        return strip_null_byte(sql).strip(";")

    def return_table(self, sql_answer: SQLAnswer, row_descs=True):
        fields = self.to_postgres_fields(sql_answer.result_set.columns)
        rows = self.to_postgres_rows(sql_answer.result_set)
        if row_descs:
            self.send(RowDescriptions(fields=fields))
        self.send(DataRow(rows=rows))
        encoding = self.get_encoding()
        tag = ("SELECT %s" % str(len(rows))).encode(encoding)
        self.send(CommandComplete(tag=tag))
        return True

    def return_error(self, sql_answer: SQLAnswer):
        self.send(Error.from_answer(error_code=sql_answer.error_code, error_message=sql_answer.error_message))
        return True

    def query(self, message: Query) -> bool:
        self.logger.debug("Postgres Proxy: Got query of:\n%s" % message.sql)
        sql = message.get_parsed_sql()
        sql_answer = self.process_query(sql)
        self.respond_from_sql_answer(sql=sql, sql_answer=sql_answer)
        self.send_ready()
        return True

    def respond_from_sql_answer(self, sql, sql_answer: SQLAnswer, row_descs=True) -> bool:
        # TODO Add command complete passthrough for Complex Queries that exceed row limit in one go
        rows = 0
        if sql_answer.result_set:
            rows = len(sql_answer.result_set)
        if RESPONSE_TYPE.OK == sql_answer.type:
            return self.return_ok(sql, rows=rows)
        elif RESPONSE_TYPE.TABLE == sql_answer.type:
            return self.return_table(sql_answer, row_descs=row_descs)
        elif RESPONSE_TYPE.ERROR == sql_answer.type:
            return self.return_error(sql_answer)

    @staticmethod
    def to_postgres_fields(columns: Iterable[Dict[str, Any]]) -> Sequence[PostgresField]:
        fields = []
        i = 0
        for column in columns:
            fields.append(GenericField(name=column["name"], object_id=column["type"].value, column_id=i))
            i += 1
        return fields

    def to_postgres_rows(self, rows: Iterable[Iterable[Any]]) -> Sequence[Sequence[bytes]]:
        p_rows = []
        for row in rows:
            p_row = []
            for column in row:
                if column is None:
                    column = ""
                elif type(column) == int or type(column) == float:
                    column = str(column)
                elif type(column) == list or type(column) == dict:
                    column = json.dumps(column)
                if isinstance(column, datetime.date) or isinstance(column, datetime.datetime):
                    try:
                        column = datetime.datetime.strftime(column, "%Y-%m-%d")
                    except ValueError:
                        try:
                            column = datetime.datetime.strftime(column, "%Y-%m-%dT%H:%M:%S")
                        except ValueError:
                            column = datetime.datetime.strptime(column, "%Y-%m-%dT%H:%M:%S.%f")
                if isinstance(column, bool):
                    if column:
                        column = "true"
                    else:
                        column = "false"
                p_row.append(column.encode(encoding=self.charset))
            p_rows.append(p_row)
        return p_rows

    def send_initial_data(self):
        server_encoding = self.charset.encode(self.charset)
        client_encoding = self.user_parameters.get(b"client_encoding", server_encoding)
        # TODO: Send BackendKeyData Here (55.2.1)
        self.send(ParameterStatus(name=b"server_version", value=b"14.6"))
        self.send(ParameterStatus(name=b"server_encoding", value=server_encoding))
        self.send(ParameterStatus(name=b"client_encoding", value=client_encoding))
        # TODO Send Parameters to complete set on 55.2.7 Asynchronous Operations E.G. - At present there is a
        #  hard-wired set of parameters for which ParameterStatus will be generated: they are server_version,
        #  server_encoding, client_encoding, application_name, default_transaction_read_only, in_hot_standby,
        #  is_superuser, session_authorization, DateStyle, IntervalStyle, TimeZone, integer_datetimes,
        #  and standard_conforming_strings
        return

    def send_ready(self):
        self.logger.debug("Ready for Query")
        self.send(ReadyForQuery(transaction_status=self.transaction_status))

    def main_loop(self):
        self.send_ready()
        while True:
            message: PostgresMessage = self.client_buffer.read_message()
            if message is None:  # Empty Data, Buffer done
                break
            tof = type(message)
            if tof in self.message_map:
                res = self.message_map[tof](message)
                if not res:
                    break
            else:
                self.logger.warning("Ignoring unsupported message type %s" % tof)

    @staticmethod
    def startProxy():
        host = config["api"]["postgres"]["host"]
        port = int(config["api"]["postgres"]["port"])
        server = TcpServer((host, port), PostgresProxyHandler)
        server.connection_id = 0
        server.mindsdb_config = config
        server.check_auth = partial(check_auth, config=config)
        server.serve_forever()


class TcpServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True


if __name__ == "__main__":
    PostgresProxyHandler.startProxy()



import json
from typing import Any, AsyncIterable, Dict, List
import requests
import httpx
from mindsdb.api.a2a.utils import to_serializable, convert_a2a_message_to_qa_format
from mindsdb.api.a2a.constants import DEFAULT_STREAM_TIMEOUT
from mindsdb.utilities import log
from mindsdb.utilities.config import config

logger = log.getLogger(__name__)


class MindsDBAgent:
    """An agent that communicates with MindsDB over HTTP following the A2A protocol."""

    SUPPORTED_CONTENT_TYPES = ["text", "text/plain", "application/json"]

    def __init__(
        self,
        agent_name="my_agent",
        project_name="mindsdb",
        user_info: Dict[str, Any] = None,
    ):
        self.agent_name = agent_name
        self.project_name = project_name
        port = config.get("api", {}).get("http", {}).get("port", 47334)
        self.base_url = f"http://localhost:{port}"
        self.agent_url = f"{self.base_url}/api/projects/{project_name}/agents/{agent_name}"
        self.sql_url = f"{self.base_url}/api/sql/query"
        self.headers = {k: v for k, v in user_info.items() if v is not None} or {}
        logger.info(f"Initialized MindsDB agent connector to {self.base_url}")

    def invoke(self, query, session_id) -> Dict[str, Any]:
        """Send a query to the MindsDB agent using SQL API."""
        try:
            escaped_query = query.replace("'", "''")
            sql_query = f"SELECT * FROM {self.project_name}.{self.agent_name} WHERE question = '{escaped_query}'"
            logger.debug(f"Sending SQL query to MindsDB: {sql_query[:100]}...")
            response = requests.post(self.sql_url, json={"query": sql_query}, headers=self.headers)
            response.raise_for_status()
            data = response.json()
            logger.debug(f"Received response from MindsDB: {json.dumps(data)[:200]}...")
            if "data" in data and len(data["data"]) > 0:
                result_row = data["data"][0]
                for column in ["response", "result", "answer", "completion", "output"]:
                    if column in result_row:
                        content = result_row[column]
                        logger.info(f"Found result in column '{column}': {content[:100]}...")
                        return {
                            "content": content,
                            "parts": [{"type": "text", "text": content}],
                        }
                logger.info("No specific result column found, returning full row")
                content = json.dumps(result_row, indent=2)
                parts = [{"type": "text", "text": content}]
                if isinstance(result_row, dict):
                    parts.append(
                        {
                            "type": "data",
                            "data": result_row,
                            "metadata": {"subtype": "json"},
                        }
                    )
                return {
                    "content": content,
                    "parts": parts,
                }
            else:
                error_msg = "Error: No data returned from MindsDB"
                logger.error(error_msg)
                return {
                    "content": error_msg,
                    "parts": [{"type": "text", "text": error_msg}],
                }
        except requests.exceptions.RequestException as e:
            error_msg = f"Error connecting to MindsDB: {str(e)}"
            logger.error(error_msg)
            return {
                "content": error_msg,
                "parts": [{"type": "text", "text": error_msg}],
            }
        except Exception as e:
            error_msg = f"Error: {str(e)}"
            logger.error(error_msg)
            return {
                "content": error_msg,
                "parts": [{"type": "text", "text": error_msg}],
            }

    async def streaming_invoke(self, messages, timeout=DEFAULT_STREAM_TIMEOUT):
        url = f"{self.base_url}/api/projects/{self.project_name}/agents/{self.agent_name}/completions/stream"
        logger.debug(f"Sending streaming request to MindsDB agent: {self.agent_name}")
        async with httpx.AsyncClient(timeout=timeout, headers=self.headers) as client:
            async with client.stream("POST", url, json={"messages": to_serializable(messages)}) as response:
                response.raise_for_status()
                async for line in response.aiter_lines():
                    if not line.strip():
                        continue
                    # Only process actual SSE data lines
                    if line.startswith("data:"):
                        payload = line[len("data:") :].strip()
                        try:
                            yield json.loads(payload)
                        except Exception as e:
                            logger.error(f"Failed to parse SSE JSON payload: {e}; line: {payload}")
                    # Ignore comments or control lines
                # Signal the end of the stream
                yield {"is_task_complete": True}

    async def stream(
        self,
        query: str,
        session_id: str,
        history: List[dict] | None = None,
        timeout: int = DEFAULT_STREAM_TIMEOUT,
    ) -> AsyncIterable[Dict[str, Any]]:
        """Stream responses from the MindsDB agent (uses streaming API endpoint)."""
        try:
            # Create A2A message structure with history and current query
            a2a_message = {"role": "user", "parts": [{"text": query}]}
            if history:
                a2a_message["history"] = history
            # Convert to Q&A format using centralized utility
            formatted_messages = convert_a2a_message_to_qa_format(a2a_message)
            logger.debug(f"Formatted messages for agent: {formatted_messages}")
            streaming_response = self.streaming_invoke(formatted_messages, timeout=timeout)
            async for chunk in streaming_response:
                content_value = chunk.get("text") or chunk.get("output") or json.dumps(chunk)
                wrapped_chunk = {"is_task_complete": False, "content": content_value, "metadata": {}}
                yield wrapped_chunk
        except Exception as e:
            logger.error(f"Error in streaming: {str(e)}")
            yield {
                "is_task_complete": True,
                "parts": [
                    {
                        "type": "text",
                        "text": f"Error: {str(e)}",
                    }
                ],
                "metadata": {
                    "type": "reasoning",
                    "subtype": "error",
                },
            }



from typing import Any, List, Optional
import ast as py_ast

import pandas as pd
from mindsdb_sql_parser.ast import ASTNode, Select, Insert, Update, Delete, Star
from mindsdb_sql_parser.ast.select.identifier import Identifier

from mindsdb.integrations.utilities.sql_utils import (
    extract_comparison_conditions,
    filter_dataframe,
    sort_dataframe,
    FilterCondition,
    FilterOperator,
    SortColumn,
)
from mindsdb.integrations.libs.base import BaseHandler
from mindsdb.integrations.libs.api_handler_exceptions import TableAlreadyExists, TableNotFound

from mindsdb.integrations.libs.response import HandlerResponse as Response, RESPONSE_TYPE
from mindsdb.utilities import log


logger = log.getLogger("mindsdb")


class FuncParser:
    def from_string(self, query_string):
        body = py_ast.parse(query_string.strip(), mode="eval").body

        if not isinstance(body, py_ast.Call):
            raise RuntimeError(f"Api function not found {query_string}")

        fnc_name = body.func.id

        params = {}
        for keyword in body.keywords:
            name = keyword.arg
            value = self.process(keyword.value)

            params[name] = value

        return fnc_name, params

    def process(self, node):
        if isinstance(node, py_ast.List):
            elements = []
            for node2 in node.elts:
                elements.append(self.process(node2))
            return elements

        if isinstance(node, py_ast.Dict):
            keys = []
            for node2 in node.keys:
                if isinstance(node2, py_ast.Constant):
                    value = node2.value
                elif isinstance(node2, py_ast.Str):  # py37
                    value = node2.s
                else:
                    raise NotImplementedError(f"Unknown dict key {node2}")

                keys.append(value)

            values = []
            for node2 in node.values:
                values.append(self.process(node2))

            return dict(zip(keys, values))

        if isinstance(node, py_ast.Name):
            # special attributes
            name = node.id
            if name == "true":
                return True
            elif name == "false":
                return False
            elif name == "null":
                return None

        if isinstance(node, py_ast.Constant):
            return node.value

        # ---- python 3.7 objects -----
        if isinstance(node, py_ast.Str):
            return node.s

        if isinstance(node, py_ast.Num):
            return node.n

        # -----------------------------

        if isinstance(node, py_ast.UnaryOp):
            if isinstance(node.op, py_ast.USub):
                value = self.process(node.operand)
                return -value

        raise NotImplementedError(f"Unknown node {node}")


class APITable:
    def __init__(self, handler):
        self.handler = handler

    def select(self, query: Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """

        raise NotImplementedError()

    def insert(self, query: Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        raise NotImplementedError()

    def update(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        raise NotImplementedError()

    def delete(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Delete

        Returns:
            None
        """
        raise NotImplementedError()

    def get_columns(self) -> list:
        """Maps the columns names from the API call resource

        Returns:
            List
        """
        raise NotImplementedError()


class APIResource(APITable):
    def __init__(self, *args, table_name=None, **kwargs):
        self.table_name = table_name
        super().__init__(*args, **kwargs)

    def select(self, query: Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """

        conditions = self._extract_conditions(query.where)

        limit = None
        if query.limit:
            limit = query.limit.value

        sort = None
        if query.order_by and len(query.order_by) > 0:
            sort = []
            for an_order in query.order_by:
                sort.append(SortColumn(an_order.field.parts[-1], an_order.direction.upper() != "DESC"))

        targets = []
        for col in query.targets:
            if isinstance(col, Identifier):
                targets.append(col.parts[-1])

        kwargs = {"conditions": conditions, "limit": limit, "sort": sort, "targets": targets}
        if self.table_name is not None:
            kwargs["table_name"] = self.table_name

        result = self.list(**kwargs)

        filters = []
        for cond in conditions:
            if not cond.applied:
                filters.append([cond.op.value, cond.column, cond.value])

        result = filter_dataframe(result, filters)

        if sort:
            sort_columns = []
            for idx, a_sort in enumerate(sort):
                if not a_sort.applied:
                    sort_columns.append(query.order_by[idx])

            result = sort_dataframe(result, sort_columns)

        if limit is not None and len(result) > limit:
            result = result[: int(limit)]

        return result

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs,
    ):
        """
        List items based on specified conditions, limits, sorting, and targets.

        Args:
            conditions (List[FilterCondition]): Optional. A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.
            limit (int): Optional. An integer to limit the number of items to be listed.
            sort (List[SortColumn]): Optional. A list of sorting criteria
            targets (List[str]): Optional. A list of strings representing specific fields

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def insert(self, query: Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """

        columns = [col.name for col in query.columns]

        data = [dict(zip(columns, a_row)) for a_row in query.values]
        kwargs = {}
        if self.table_name is not None:
            kwargs["table_name"] = self.table_name

        self.add(data, **kwargs)

    def add(self, row: List[dict], **kwargs) -> None:
        """
        Add a single item to the dataa collection

        Args:
        r   ow (dict): A dictionary representing the item to be added.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def update(self, query: Update) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update

        Returns:
            None
        """
        conditions = self._extract_conditions(query.where)

        values = {key: val.value for key, val in query.update_columns.items()}

        self.modify(conditions, values)

    def modify(self, conditions: List[FilterCondition], values: dict):
        """
        Modify items based on specified conditions and values.

        Args:
            conditions (List[FilterCondition]): A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.
            values (dict): A dictionary of values to be updated.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError

    def delete(self, query: Delete) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Delete

        Returns:
            None
        """
        conditions = self._extract_conditions(query.where)

        self.remove(conditions)

    def remove(self, conditions: List[FilterCondition]):
        """
        Remove items based on specified conditions.

        Args:
            conditions (List[FilterCondition]): A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def _extract_conditions(self, where: ASTNode) -> List[FilterCondition]:
        return [FilterCondition(i[1], FilterOperator(i[0].upper()), i[2]) for i in extract_comparison_conditions(where)]


class MetaAPIResource(APIResource):
    # TODO: Add a meta_table_info() method in case metadata cannot be retrieved as expected below?

    def meta_get_tables(self, table_name: str, **kwargs) -> dict:
        """
        Retrieves table metadata for the API resource.

        Args:
            table_name (str): The name given to the table that represents the API resource. This is required because the name for the APIResource is given by the handler.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            Dict: The dictionary should contain the following fields:
            - TABLE_NAME (str): Name of the table.
            - TABLE_TYPE (str): Type of the table, e.g. 'BASE TABLE', 'VIEW', etc. (optional).
            - TABLE_SCHEMA (str): Schema of the table (optional).
            - TABLE_DESCRIPTION (str): Description of the table (optional).
            - ROW_COUNT (int): Estimated number of rows in the table (optional).
        """
        pass

    def meta_get_columns(self, table_name: str, **kwargs) -> List[dict]:
        """
        Retrieves column metadata for the API resource.

        Args:
            table_name (str): The name given to the table that represents the API resource. This is required because the name for the APIResource is given by the handler.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            List[dict]: The list should contain dictionaries with the following fields:
            - TABLE_NAME (str): Name of the table.
            - COLUMN_NAME (str): Name of the column.
            - DATA_TYPE (str): Data type of the column, e.g. 'VARCHAR', 'INT', etc.
            - COLUMN_DESCRIPTION (str): Description of the column (optional).
            - IS_NULLABLE (bool): Whether the column can contain NULL values (optional).
            - COLUMN_DEFAULT (str): Default value of the column (optional).
        """
        pass

    def meta_get_column_statistics(self, table_name: str, **kwargs) -> List[dict]:
        """
        Retrieves column statistics for the API resource.

        Args:
            table_name (str): The name given to the table that represents the API resource. This is required because the name for the APIResource is given by the handler.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            List[dict]: The list should contain dictionaries with the following fields:
            - TABLE_NAME (str): Name of the table.
            - COLUMN_NAME (str): Name of the column.
            - MOST_COMMON_VALUES (List[str]): Most common values in the column (optional).
            - MOST_COMMON_FREQUENCIES (List[str]): Frequencies of the most common values in the column (optional).
            - NULL_PERCENTAGE: Percentage of NULL values in the column (optional).
            - MINIMUM_VALUE (str): Minimum value in the column (optional).
            - MAXIMUM_VALUE (str): Maximum value in the column (optional).
            - DISTINCT_VALUES_COUNT (int): Count of distinct values in the column (optional).
        """
        pass

    def meta_get_primary_keys(self, table_name: str, **kwargs) -> List[dict]:
        """
        Retrieves primary key metadata for the API resource.

        Args:
            table_name (str): The name given to the table that represents the API resource. This is required because the name for the APIResource is given by the handler.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            List[dict]: The list should contain dictionaries with the following fields:
            - TABLE_NAME (str): Name of the table.
            - COLUMN_NAME (str): Name of the column that is part of the primary key.
            - ORDINAL_POSITION (int): Position of the column in the primary key (optional).
            - CONSTRAINT_NAME (str): Name of the primary key constraint (optional).
        """
        pass

    def meta_get_foreign_keys(self, table_name: str, all_tables: List[str], **kwargs) -> List[dict]:
        """
        Retrieves foreign key metadata for the API resource.

        Args:
            table_name (str): The name given to the table that represents the API resource. This is required because the name for the APIResource is given by the handler.
            all_tables (List[str]): A list of all table names in the API resource. This is used to identify relationships between tables.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            List[dict]: The list should contain dictionaries with the following fields:
            - PARENT_TABLE_NAME (str): Name of the parent table.
            - PARENT_COLUMN_NAME (str): Name of the parent column that is part of the foreign key.
            - CHILD_TABLE_NAME (str): Name of the child table.
            - CHILD_COLUMN_NAME (str): Name of the child column that is part of the foreign key.
            - CONSTRAINT_NAME (str): Name of the foreign key constraint (optional).
        """
        pass


class APIHandler(BaseHandler):
    """
    Base class for handlers associated to the applications APIs (e.g. twitter, slack, discord  etc.)
    """

    def __init__(self, name: str):
        super().__init__(name)
        """ constructor
        Args:
            name (str): the handler name
        """
        self._tables = {}

    def _register_table(self, table_name: str, table_class: Any):
        """
        Register the data resource. For e.g if you are using Twitter API it registers the `tweets` resource from `/api/v2/tweets`.
        """
        if table_name.lower() in self._tables:
            raise TableAlreadyExists(f"Table with name {table_name} already exists for this handler")
        self._tables[table_name.lower()] = table_class

    def _get_table(self, name: Identifier):
        """
        Check if the table name was added to the _register_table
        Args:
            name (Identifier): the table name
        """
        name = name.parts[-1].lower()
        if name in self._tables:
            return self._tables[name]
        raise TableNotFound(f"Table not found: {name}")

    def query(self, query: ASTNode):
        if isinstance(query, Select):
            # If the list method exists, it should be overridden in the child class.
            # The APIResource class could be used as a base class by overriding the select method, but not the list method.
            table = self._get_table(query.from_table)
            list_method = getattr(table, "list", None)
            if not list_method or (list_method and list_method.__func__ is APIResource.list):
                # for back compatibility, targets wasn't passed in previous version
                query.targets = [Star()]
            result = self._get_table(query.from_table).select(query)
        elif isinstance(query, Update):
            result = self._get_table(query.table).update(query)
        elif isinstance(query, Insert):
            result = self._get_table(query.table).insert(query)
        elif isinstance(query, Delete):
            result = self._get_table(query.table).delete(query)
        else:
            raise NotImplementedError

        if result is None:
            return Response(RESPONSE_TYPE.OK)
        elif isinstance(result, pd.DataFrame):
            return Response(RESPONSE_TYPE.TABLE, result)
        else:
            raise NotImplementedError

    def get_columns(self, table_name: str) -> Response:
        """
        Returns a list of entity columns
        Args:
            table_name (str): the table name
        Returns:
            RESPONSE_TYPE.TABLE
        """

        result = self._get_table(Identifier(table_name)).get_columns()

        df = pd.DataFrame(result, columns=["Field"])
        df["Type"] = "str"

        return Response(RESPONSE_TYPE.TABLE, df)

    def get_tables(self) -> Response:
        """
        Return list of entities
        Returns:
            RESPONSE_TYPE.TABLE
        """
        result = list(self._tables.keys())

        df = pd.DataFrame(result, columns=["table_name"])
        df["table_type"] = "BASE TABLE"

        return Response(RESPONSE_TYPE.TABLE, df)


class MetaAPIHandler(APIHandler):
    """
    Base class for handlers associated to the applications APIs (e.g. twitter, slack, discord  etc.)

    This class is used when the handler is also needed to store information in the data catalog.
    """

    def meta_get_handler_info(self, **kwargs) -> str:
        """
        Retrieves information about the design and implementation of the API handler.
        This should include, but not be limited to, the following:
        - The type of SQL queries and operations that the handler supports.
        - etc.

        Args:
            kwargs: Additional keyword arguments that may be used in generating the handler information.

        Returns:
            str: A string containing information about the API handler's design and implementation.
        """
        pass

    def meta_get_tables(self, table_names: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Retrieves metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve metadata.
            kwargs: Additional keyword arguments that may be used by the specific API resource implementation.

        Returns:
            Response: A response object containing the table metadata.
        """
        df = pd.DataFrame()
        for table_name, table_class in self._tables.items():
            if table_names is None or table_name in table_names:
                try:
                    if hasattr(table_class, "meta_get_tables"):
                        table_metadata = table_class.meta_get_tables(table_name, **kwargs)
                        df = pd.concat([df, pd.DataFrame([table_metadata])], ignore_index=True)
                except Exception as e:
                    logger.error(f"Error retrieving metadata for table {table_name}: {e}")

        return Response(RESPONSE_TYPE.TABLE, df)

    def meta_get_columns(self, table_names: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Retrieves column metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve column metadata.

        Returns:
            Response: A response object containing the column metadata.
        """
        df = pd.DataFrame()
        for table_name, table_class in self._tables.items():
            if table_names is None or table_name in table_names:
                try:
                    if hasattr(table_class, "meta_get_columns"):
                        column_metadata = table_class.meta_get_columns(table_name, **kwargs)
                        df = pd.concat([df, pd.DataFrame(column_metadata)], ignore_index=True)
                except Exception as e:
                    logger.error(f"Error retrieving column metadata for table {table_name}: {e}")

        return Response(RESPONSE_TYPE.TABLE, df)

    def meta_get_column_statistics(self, table_names: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Retrieves column statistics for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve column statistics.

        Returns:
            Response: A response object containing the column statistics.
        """
        df = pd.DataFrame()
        for table_name, table_class in self._tables.items():
            if table_names is None or table_name in table_names:
                try:
                    if hasattr(table_class, "meta_get_column_statistics"):
                        column_statistics = table_class.meta_get_column_statistics(table_name, **kwargs)
                        df = pd.concat([df, pd.DataFrame(column_statistics)], ignore_index=True)
                except Exception as e:
                    logger.error(f"Error retrieving column statistics for table {table_name}: {e}")

        return Response(RESPONSE_TYPE.TABLE, df)

    def meta_get_primary_keys(self, table_names: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Retrieves primary key metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve primary key metadata.

        Returns:
            Response: A response object containing the primary key metadata.
        """
        df = pd.DataFrame()
        for table_name, table_class in self._tables.items():
            if table_names is None or table_name in table_names:
                try:
                    if hasattr(table_class, "meta_get_primary_keys"):
                        primary_key_metadata = table_class.meta_get_primary_keys(table_name, **kwargs)
                        df = pd.concat([df, pd.DataFrame(primary_key_metadata)], ignore_index=True)
                except Exception as e:
                    logger.error(f"Error retrieving primary keys for table {table_name}: {e}")

        return Response(RESPONSE_TYPE.TABLE, df)

    def meta_get_foreign_keys(self, table_names: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Retrieves foreign key metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve foreign key metadata.

        Returns:
            Response: A response object containing the foreign key metadata.
        """
        df = pd.DataFrame()
        all_tables = list(self._tables.keys())
        for table_name, table_class in self._tables.items():
            if table_names is None or table_name in table_names:
                try:
                    if hasattr(table_class, "meta_get_foreign_keys"):
                        foreign_key_metadata = table_class.meta_get_foreign_keys(
                            table_name, all_tables=table_names if table_names else all_tables, **kwargs
                        )
                        df = pd.concat([df, pd.DataFrame(foreign_key_metadata)], ignore_index=True)
                except Exception as e:
                    logger.error(f"Error retrieving foreign keys for table {table_name}: {e}")

        return Response(RESPONSE_TYPE.TABLE, df)


class APIChatHandler(APIHandler):
    def get_chat_config(self):
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()

    def get_my_user_name(self) -> list:
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()
